##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import os

def makedirs(path: str):
    """
    Creates a directory and all intermediate directories if they do not exist in a file path.

    Parameters:
        path (str): The directory path to create.
    """

    if not os.path.isdir(path):
        os.makedirs(path)


def makedirs_for_file(filepath: str):
    """
    Creates the necessary directories for a given file path if they do not already exist.

    Parameters:
        filepath (str): The complete file path for which the directories are to be created.
    """

    directory_path, _ = os.path.split(filepath)
    makedirs(directory_path)




##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
from .numpy import load_numpy_files
from .numpy import save_dict_to_numpy_files

from .general import makedirs
from .general import makedirs_for_file

from .odsreader import ODSReader

from .json import ExputilsJSONEncoder
from .json import exputils_json_object_hook
from .json import save_dict_as_json_file
from .json import load_dict_from_json_file
from .json import convert_json_dict_keys_to_ints

from .dill import load_dill
from .dill import save_dill
from .dill import load_dill_files



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import os
import stat
import re
import copy
import shutil
from typing import Optional
import exputils
from collections import OrderedDict

# TODO: allow to delete a line in the config, for example if the value of a param is "#RM"


def generate_experiment_files(ods_filepath: Optional[str] = None,
                              directory: Optional[str] = None,
                              extra_files: Optional[list] = None,
                              extra_experiment_files: Optional[list] = None,
                              verbose: bool = False,
                              copy_operator: str = 'shutil'):
    """
    Generates experiments based on a configuration ODS file (LibreOffice Spreadsheet) and template
    source code.

    The configuration ODS has to be in a specific form.
    See `resources/experiment_configurations.ods` for an example file.

    The template source code is usually located in `.\src\exp` for code on the experiment level
    and `.\src\\rep` for code on the repetition level.

    [//]: # (TODO: either remove or document the options for extra-files)

    Parameters:
        ods_filepath (str):
            Path to the ODS configuration file that defines the experiments.
            Default is `'./experiment_configurations.ods'`
        directory (str):
            Path to directory where the experiments will be generated.
            Default is `'./experiments'`.
        verbose (bool):
            Should verbose output with more information given. Default is `False`.
        copy_operator (str):
            Define the copy operator for source code files. Either 'shutil' (default) for the python
            copy function or 'cp' for the linux terminal cp operator. The choice of the 'cp' copy
            operator was introduced as for some OS systems the 'shutil' did not work under python 3.8.

    Notes:

    - Sheets in the configuration ODS file define groups of experiments for which an extra
       subfolder in the output directory will be generated.
    """

    if ods_filepath is None:
        ods_filepath = os.path.join('.', exputils.DEFAULT_ODS_CONFIGURATION_FILE)

    if directory is None:
        directory = os.path.join('.',exputils. DEFAULT_EXPERIMENTS_DIRECTORY)
    elif directory == '':
        directory = '.'

    if verbose:
        print('Load config from {!r} ...'.format(ods_filepath))

    config_data = _load_configuration_data_from_ods(ods_filepath)

    # generate experiment files based on the loaded configurations
    if verbose:
        print('Generate experiments ...'.format(ods_filepath))

    _generate_files_from_config(
        config_data, directory,
        extra_files=extra_files, extra_experiment_files=extra_experiment_files, verbose=verbose, copy_operator=copy_operator
    )


def _load_configuration_data_from_ods(ods_filepath):
    """

    :param ods_filepath:
    :return:
    """

    config_data = []

    doc = exputils.io.ODSReader(ods_filepath, clonespannedcolumns=False)
    for sheet_name, sheet_data in doc.sheets.items():

        experiments_data = dict()

        experiments_data['directory'] = sheet_name
        experiments_data['experiments'] = OrderedDict()

        file_config = []

        parameters_start_idx = None
        parameters_end_idx = None

        file_borders = []

        variable_names = []

        #######################################################
        # first row: search for columns that describe parameters
        row_idx = 0

        for col_idx, col_data in enumerate(sheet_data[row_idx]):
            if col_data is not None:
                if col_data.lower() == 'parameters':
                    if parameters_start_idx is None:
                        parameters_start_idx = col_idx
                else:
                    parameters_end_idx = col_idx - 1
                    break

        #######################################################
        # second row: identify the template files
        row_idx = 1

        col_idx = parameters_start_idx
        stop = False

        previous_template_file_path = ''
        while not stop:

            if sheet_data[row_idx][col_idx] is not None and sheet_data[row_idx][col_idx] != previous_template_file_path:
                file_config.append(dict())
                file_config[-1]['template_file_path'] = sheet_data[row_idx][col_idx]

                previous_template_file_path = sheet_data[row_idx][col_idx]

                if file_borders:
                    file_borders[-1][1] = col_idx-1
                file_borders.append([col_idx, None])

            col_idx += 1
            if col_idx >= len(sheet_data[row_idx]) or (parameters_end_idx is not None and col_idx > parameters_end_idx):
                stop = True

        # in case the first line gives a definitive end, but not the second, then use the definitive end of the first line
        if parameters_end_idx is not None and file_borders[-1][1] is None:
            file_borders[-1][1] = parameters_end_idx

        #######################################################
        # third row: identify the file name templates
        row_idx = 2

        for file_idx in range(len(file_config)):
            file_config[file_idx]['file_name_template'] = sheet_data[row_idx][file_borders[file_idx][0]]

        #######################################################
        # fourth row: identify the variable names and if there are repetitions and a experiment folder
        row_idx = 3

        # find repetitions column
        repetitions_info_col_idx = None
        for col_idx in range(parameters_start_idx):
            if sheet_data[row_idx][col_idx] is not None and sheet_data[row_idx][col_idx].lower() == 'repetitions':
                repetitions_info_col_idx = col_idx
                break

        # find column with sources for repetitions
        repetition_source_file_locations_col_idx = None
        for col_idx in range(parameters_start_idx):
            if sheet_data[row_idx][col_idx] is not None and (sheet_data[row_idx][col_idx].lower() == 'files' or sheet_data[row_idx][col_idx].lower() == 'repetition files'):
                repetition_source_file_locations_col_idx = col_idx
                break
        # if no src directory for repetitions is given, then see if src/rep exists
        default_repetition_source_folder = None
        if not repetition_source_file_locations_col_idx:
            if os.path.isdir('./src/rep'):
                default_repetition_source_folder = './src/rep'

        # find column with sources for experiments
        experiment_source_file_locations_col_idx = None
        for col_idx in range(parameters_start_idx):
            if sheet_data[row_idx][col_idx] is not None and sheet_data[row_idx][col_idx].lower() == 'experiment files':
                experiment_source_file_locations_col_idx = col_idx
                break
        # if no src directory for repetitions is given, then see if src/rep exists
        default_experiment_source_folder = None
        if not experiment_source_file_locations_col_idx:
            if os.path.isdir('./src/exp'):
                default_experiment_source_folder = './src/exp'

        # find variable names
        for file_idx in range(len(file_config)):

            start_idx = file_borders[file_idx][0]
            end_idx = file_borders[file_idx][1]

            cur_variable_names = []

            col_idx = start_idx
            stop = False
            while not stop:
                cur_variable_names.append(sheet_data[row_idx][col_idx])

                col_idx += 1

                if col_idx < len(sheet_data[row_idx]) and sheet_data[row_idx][col_idx] is None:
                    file_borders[file_idx][1] = col_idx-1
                    stop = True

                elif col_idx >= len(sheet_data[row_idx]) or (end_idx is not None and col_idx > end_idx):
                    stop = True

            variable_names.append(cur_variable_names)

        #######################################################
        # remaining rows: experiments

        for row_idx in range(4, len(sheet_data)):

            experiment_id = sheet_data[row_idx][0]

            if experiment_id is not None:
                experiment_id = int(experiment_id)

                experiments_data['experiments'][experiment_id] = dict()

                experiments_data['experiments'][experiment_id]['files'] = copy.deepcopy(file_config)

                # repetitions info if it exists
                if repetitions_info_col_idx is not None and sheet_data[row_idx][repetitions_info_col_idx] is not None:
                    experiments_data['experiments'][experiment_id]['repetitions'] = int(sheet_data[row_idx][repetitions_info_col_idx])
                else:
                    experiments_data['experiments'][experiment_id]['repetitions'] = None

                if repetition_source_file_locations_col_idx is not None and sheet_data[row_idx][repetition_source_file_locations_col_idx] is not None:
                    experiments_data['experiments'][experiment_id]['repetition_source_file_locations'] = [i.strip() for i in sheet_data[row_idx][repetition_source_file_locations_col_idx].split(',')]
                elif default_repetition_source_folder:
                    experiments_data['experiments'][experiment_id]['repetition_source_file_locations'] = [default_repetition_source_folder]
                else:
                    experiments_data['experiments'][experiment_id]['repetition_source_file_locations'] = None

                if experiment_source_file_locations_col_idx is not None and sheet_data[row_idx][experiment_source_file_locations_col_idx] is not None:
                    experiments_data['experiments'][experiment_id]['experiment_source_file_locations'] = [i.strip() for i in sheet_data[row_idx][experiment_source_file_locations_col_idx].split(',')]
                elif default_experiment_source_folder:
                    experiments_data['experiments'][experiment_id]['experiment_source_file_locations'] = [default_experiment_source_folder]
                else:
                    experiments_data['experiments'][experiment_id]['experiment_source_file_locations'] = None

                for file_idx in range(len(file_config)):

                    col_idx = file_borders[file_idx][0]

                    experiments_data['experiments'][experiment_id]['files'][file_idx]['variables'] = dict()

                    for variable_name in variable_names[file_idx]:

                        if col_idx > len(sheet_data[row_idx]) - 1:
                            # if there is no content in the final cells, the array that holds the data is shorter
                            cur_cell_data = None
                        else:
                            cur_cell_data = get_cell_data(sheet_data[row_idx][col_idx])

                        experiments_data['experiments'][experiment_id]['files'][file_idx]['variables'][variable_name] = cur_cell_data

                        col_idx +=1

        config_data.append(experiments_data)

    return config_data


def get_cell_data(data):

    if data is None:
        return None

    # replace strange characters that are not used for python strings
    data = data.replace('’', '\'')
    data = data.replace('`', '\'')

    return data


def _generate_files_from_config(config_data, directory='.', extra_files=None, extra_experiment_files=None, verbose=False, copy_operator='shutil'):
    """

    Format of configuration data:

    config_data['directory']
    config_data['experiments']: dictionary with keys=experiment id, values=description:


    config_data['ids']: Ids of the experiments
    config_data['files']:  List with configuration for each template file
        template_file['template_file_path']: filepath of the template file
        template_file['file_name_template']: template for the filename
        template_file['variables']: Dictionary with key=variable name, value=variable value

    :param config_data:
    :return:
    """

    if extra_files is None:
        extra_files = []
    elif not isinstance(extra_files, list):
        extra_files = [extra_files]

    if extra_experiment_files is None:
        extra_experiment_files = []
    elif not isinstance(extra_experiment_files, list):
        extra_experiment_files = [extra_experiment_files]

    for experiment_group_config in config_data:

        # only create group folder if more than one sheet or the sheet name is not empty or 'Sheet1'
        if len(config_data) == 1 and experiment_group_config['directory'] in ['Sheet1', '']:
            group_directory = directory
        else:
            group_directory = os.path.join(directory, experiment_group_config['directory'])

        # create the folder if not exists
        if not os.path.isdir(group_directory):
            os.makedirs(group_directory)

        # generate the experiment folders and files
        for experiment_id, experiment_config in experiment_group_config['experiments'].items():

            experiment_directory = os.path.join(group_directory, exputils.EXPERIMENT_DIRECTORY_TEMPLATE.format(experiment_id))

            # create folders for the repetitions if necessary:
            if experiment_config['repetitions'] is None:
                num_of_repetitions = 1
            else:
                num_of_repetitions = experiment_config['repetitions']

            for repetition_id in range(num_of_repetitions):

                # only create repetition folders if repetitions are specified
                if experiment_config['repetitions'] is None:
                    experiment_files_directory = experiment_directory
                else:
                    experiment_files_directory = os.path.join(experiment_directory, exputils.REPETITION_DIRECTORY_TEMPLATE.format(repetition_id))

                # create folder if not exists
                if not os.path.isdir(experiment_files_directory):
                    os.makedirs(experiment_files_directory)

                # generate the files for the experiment, or the repetition if they are defined
                if experiment_config['repetition_source_file_locations'] is None:
                    source_files = extra_files
                else:
                    source_files = experiment_config['repetition_source_file_locations'] + extra_files

                _generate_source_files(
                    source_files,
                    experiment_files_directory,
                    experiment_config,
                    experiment_id,
                    repetition_id,
                    copy_operator=copy_operator
                )

            # if there are experiment - repetitions defined, then generate the files for the experiment folder
            if experiment_config['experiment_source_file_locations'] is None:
                source_files = extra_experiment_files
            else:
                source_files = experiment_config['experiment_source_file_locations'] + extra_experiment_files

            if source_files:
                _generate_source_files(
                    source_files,
                    experiment_directory,
                    experiment_config,
                    experiment_id,
                    copy_operator=copy_operator
                )


def _generate_source_files(source_files, experiment_files_directory, experiment_config, experiment_id, repetition_id=None, copy_operator='shutil'):

    default_value = ''

    if copy_operator.lower() == 'shutil':
        copy_function = _copy_operator_shutil
    elif copy_operator.lower() == 'cp':
        copy_function = _copy_operator_linux_cp
    else:
        raise ValueError('Unknown copy_operator "{}"! Must be either "shutil" or "cp".'.format(copy_operator))

    # create the source files that are given by templates
    for file_config in experiment_config['files']:

        # check if the given file exists, otherwise it might be in one of the given directories under the source-files property of the experiment
        template_file_path = None
        if os.path.isfile(file_config['template_file_path']):
            template_file_path = file_config['template_file_path']
        else:
            for src in source_files:
                if os.path.isdir(src):
                    if os.path.isfile(os.path.join(src, file_config['template_file_path'])):
                        template_file_path = os.path.join(src, file_config['template_file_path'])
                        break

        if template_file_path is not None:

            # get file permissions
            permissions = os.stat(template_file_path)[stat.ST_MODE]

            # Read in the template file
            file_lines = []
            with open(template_file_path, 'r') as file:
                file_lines = file.readlines()

            # lines that should be written
            write_lines = []
            for line in file_lines:

                # Replace the variables
                line = line.replace('<experiment_id>', str(experiment_id))

                if repetition_id is not None:
                    line = line.replace('<repetition_id>', str(repetition_id))

                is_remove_line = False
                for variable_name, variable_value in file_config['variables'].items():

                    match = True
                    while match is not None:
                        # allow default values that come directly after the varibale_name: "<var_name,'varibale'>"
                        match = re.search(r"<{}(,[^<]+)?>".format(variable_name), line, flags=re.IGNORECASE)

                        if match is not None:
                            # value is the variable_value
                            val = variable_value

                            # if it is None, then use the default value from the file, or the general default value which is ''
                            if val is None:
                                # if there is a default value defined in the source file
                                if match.group(1) is None:
                                    val = default_value
                                else:
                                    val = match.group(1)[1:]  # remove the initial ','

                            # delete the whole line if the varibale value is "%RM"
                            if val == '%RM':
                                is_remove_line = True
                                break

                            line = line.replace(match.group(0), val)

                # delete the whole line if the varibale value is "%RM"
                if not is_remove_line:
                    write_lines.append(line)

            # Write the final output file
            file_path = os.path.join(experiment_files_directory, file_config['file_name_template'].format(experiment_id))
            with open(file_path, 'w') as file:
                file.writelines(write_lines)
            os.chmod(file_path, permissions)

    # copy all other sources, but not the templates if they are in one of the source directories
    template_files = [file_config['template_file_path'] for file_config in experiment_config['files']]

    for src in source_files:
        _copy_experiment_files(src, experiment_files_directory, template_files, copy_function)


def _copy_experiment_files(src, dst, template_files, copy_function):

    if os.path.isdir(src):
        # if directory, then copy the content

        for item in os.listdir(src):
            s = os.path.join(src, item)

            # if subdirectory, then delete any existing directory and make a new directory

            if os.path.isdir(s):

                d = os.path.join(dst, item)

                if os.path.isdir(d):
                    shutil.rmtree(d, ignore_errors=True)

                os.mkdir(d)

            else:
                d = dst

            _copy_experiment_files(s, d, template_files, copy_function)

    else:
        # if file, then copy it directly

        # do not copy the template files, because they were already processed
        if os.path.basename(src) not in [os.path.basename(f) for f in template_files]:
            copy_function(src, dst)


def _copy_operator_shutil(src, dst):
    shutil.copy2(src, dst)


def _copy_operator_linux_cp(src, dst):
    os.system('cp "{}" "{}"'.format(src, dst))




##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import ipywidgets
from exputils.gui.jupyter.experiment_data_selection_widget import ExperimentDataSelectionWidget
import IPython


DEFAULT_PLOTLY_MEANSTD_SCATTER_CONFIG = """layout = dict(
    xaxis = dict(
        title = '', 
        range = [None, None]),
    yaxis = dict(
        title = '', 
        range = [None, None]),
),
moving_average = dict(
    n = 1
),
data_filter = dict(
    every_nth_step = dict(
        step = 1,
        include_final_step = False
    ),
),        
default_group_label = 'rep <group_idx>'"""

DEFAULT_PLOTLY_BOX_CONFIG = """layout = dict(
    xaxis = dict(
        title = '', 
        range = [None, None] 
        ),
    yaxis = dict(
        title = '', 
        range = [None, None] 
        )
    )"""

DEFAULT_PLOTLY_MEANSTD_BAR_CONFIG = """layout = dict(
    xaxis = dict(
        title = '', 
        range = [None, None] 
        ),
    yaxis = dict(
        title = '', 
        range = [None, None] 
        )
    ),
    default_group_label = 'rep <group_idx>'"""

DEFAULT_TABULATE_MEANSTD_CONFIG = """primary_content_function = np.nanmean,
secondary_content_function = np.nanstd,
flip_rows_and_cols = False,
tabulate=eu.AttrDict(
            tablefmt='html', #
            numalign='right',
        ),
cell_format = '{:.3f} ({:.3f})',
top_left_cell_content = ''"""

DEFAULT_TABULATE_PAIRWISE_CONFIG = """pairwise_function = eu.misc.mannwhitneyu_pvalue,
pairwise_mode = 'upper_triangle', # which pairs are compared? 'full', 'full_not_identity', 'upper_triangle', 'upper_triangle_not_identity', 'lower_triangle', 'lower_triangle_not_identity'
tabulate=eu.AttrDict(
            tablefmt='html', # 'html' or 'latex'
            numalign='right',
        ),
cell_format = '{}',
top_left_cell_content = 'p-value',
labels=[]"""

CODE_TEMPLATE_MULTILINE = """# Plotting of <datasources> 
import exputils as eu
<import_statements>

plot_config = <plot_function_config>

selection_widget = eu.gui.jupyter.ExperimentDataPlotSelectionWidget(
    experiment_data_loader,
    datasources=<datasources>,
    experiment_ids='all',
    repetition_ids='all',
    output_format=<output_format>,
    data_filter=<data_filter>,
    plot_function=<plot_function>,
    plot_function_config=plot_config,
    state_backup_name=<state_backup_name>,
    state_backup_variable_filter=['experiment_ids', 'repetition_ids'],  # only save these variables as backup
    is_datasources_selection=False,
    is_output_format_selection=False,
    is_data_filter_selection=False,
    is_plot_function_selection=False,
    is_plot_function_config_editor=False,
    is_code_producer=False) 
display(selection_widget)
selection_widget.plot_data()"""

CODE_TEMPLATE_SINGLELINE = """# Plotting of <datasources> 
import exputils as eu
<import_statements>

plot_config = <plot_function_config>

selection_widget = eu.gui.jupyter.ExperimentDataPlotSelectionWidget(experiment_data_loader, datasources=<datasources>, experiment_ids='all', repetition_ids='all', output_format=<output_format>, data_filter=<data_filter>, plot_function=<plot_function>, plot_function_config=plot_config, state_backup_name=<state_backup_name>, state_backup_variable_filter=['experiment_ids', 'repetition_ids'], is_datasources_selection=False, is_output_format_selection=False, is_data_filter_selection=False, is_plot_function_selection=False, is_plot_function_config_editor=False, is_code_producer=False)
display(selection_widget)
selection_widget.plot_data()"""

def _config_obj_to_dict(config_obj):
    '''
    Transforms a potential configuration (either dict or string) to configuration dictionary.
    If it is a string, it tries to execute the <string> as python code with 'dict(<string>)'.
    If this fails, then the returned object the Exception.
    '''
    if config_obj is None:
        config = None
    elif isinstance(config_obj, dict):
        config = config_obj
    else:
        try:
            _locals = {}
            exec('import numpy as np;import exputils as eu;tmp_dict = dict({})'.format(config_obj), {}, _locals)
            config = _locals['tmp_dict']
        except Exception as err:
            config = err
    return config


class ExperimentDataPlotSelectionWidget(ExperimentDataSelectionWidget):
    """
        Jupyter widget for plotting experiment data and creating Jupyter cells for dedicated plotting.

        The widget allows to select the datasource that should be plotted and the plotting function.
        It also allows to select which experiments should be plotted and to create dedicated Jupyter
        cells to plot specific datasources.

        GUI of the widget:
        <figure markdown="span">
          ![ExperimentDataPlotSelectionWidget](../assets/images/experiment_data_plot_selection_widget.png)
        </figure>

        Functionality:

        - _Data Sources_: Allows to define the datasource or datasources that should be plotted.
            The datasource names correspond to the filenames under the data folder of repetitions and
            correspond to the names that were used by the [logging](logging.md#writting) functions.
            A comma-seperated list of datasources can be provided for table plots ([tabulate_meanstd][exputils.gui.jupyter.tabulate_meanstd.tabulate_meanstd]).
            It is also possible to extract single elements from data arrays using bracket operation after the name.
            For example `loss[-1]` will access the final loss value.
        - _Experiments_: Selection of experiments from which data was loaded that should be plotted.
        - _Repetitions_: Selection of repetitions from which data was loaded that should be plotted.
        - _Plot Function_: Plotting function that should be used.
            See the [Plotting Functions](./#plotting-functions) section for a list of exisiting
            plotting functions.
        - _Plot Configuration_: Configuration of the plotting function.
            See the [Plotting Functions](./#plotting-functions) section for details.
        - _Plot Data_: Plots the data below the widget.
        - _Code Production_: Creates a new Jupyter notebook cell below the current one that contains
            the code to plot the data again with all the configuration that was set in the GUI.
            The code also allows to change the configuration.

        Example:
            Execute the following code in a Jupyter notebook located in the experiment campaign directory under a subdirectory, such as `./analysis`.
            This code should be executed after data has been loaded, for example via the [ExperimentDataLoaderWidget][exputils.gui.jupyter.ExperimentDataLoaderWidget].
            ```python
            # allow plotting of data loaded by the experiment_data_loader (ExperimentDataLoaderWidget)
            experiment_data_plotter = eu.gui.jupyter.ExperimentDataPlotSelectionWidget(experiment_data_loader)
            display(experiment_data_plotter)
            ```
    """

    @staticmethod
    def default_config():
        dc = ExperimentDataSelectionWidget.default_config()

        # do not show the get data button, this widget will create its own plotting button
        dc.is_get_experiment_data_button = False

        # dictionary with possible plotting function
        dc.plot_functions = {'plotly_meanstd_scatter': eu.gui.jupyter.plotly_meanstd_scatter,
                             'plotly_box': eu.gui.jupyter.plotly_box,
                             'plotly_meanstd_bar': eu.gui.jupyter.plotly_meanstd_bar,
                             'tabulate_meanstd': eu.gui.jupyter.tabulate_meanstd,
                             'tabulate_pairwise': eu.gui.jupyter.tabulate_pairwise,

        }

        # dictionary with plot_function_configs for each
        dc.plot_function_configs = {'plotly_meanstd_scatter': DEFAULT_PLOTLY_MEANSTD_SCATTER_CONFIG,
                                    'plotly_box': DEFAULT_PLOTLY_BOX_CONFIG,
                                    'plotly_meanstd_bar': DEFAULT_PLOTLY_MEANSTD_BAR_CONFIG,
                                    'tabulate_meanstd': DEFAULT_TABULATE_MEANSTD_CONFIG,
                                    'tabulate_pairwise': DEFAULT_TABULATE_PAIRWISE_CONFIG,
        }

        dc.is_plot_function_selection = True
        dc.plot_function = list(dc.plot_functions.keys())[0]
        dc.plot_function_selection = eu.AttrDict(
            hbox=eu.AttrDict(  # ipywidgets.HBox parameters
                layout=eu.AttrDict(
                    width='100%')),
            label=eu.AttrDict(  # ipywidgets.Label parameters
                value='Plot Function:',
                layout=eu.AttrDict(min_width='100px')),
            dropdown=eu.AttrDict(  # ipywidgets.Text parameters
                description='',
                layout=eu.AttrDict(width='100%')))

        dc.is_plot_function_config_editor = True
        dc.plot_function_config = None
        dc.plot_function_config_editor = eu.AttrDict(
            title = 'Plot Configuration',
            accordion=eu.AttrDict(  # ipywidgets.HBox parameters
                selected_index=None,  # collapse accordion at init
                layout=eu.AttrDict(
                    width='100%')),
            textarea=eu.AttrDict(  # ipywidgets.Label parameters
                placeholder='Provide the configuration of the plot function in form of a python dictionary ...',
                layout=eu.AttrDict(
                    min_width='100%'))) # TODO: start with a larger height at the beginning

        dc.is_plot_button = True
        dc.plot_button = eu.AttrDict(
            description='Plot Data',
            tooltip='Plots the data according to the selection.',
            layout=eu.AttrDict(width='100%', height='95%'))

        dc.code_producer.code_templates = [dict(name='Multi Line', code_template=CODE_TEMPLATE_MULTILINE),
                                           dict(name='Single Line', code_template=CODE_TEMPLATE_SINGLELINE)]

        dc.figure_output = eu.AttrDict()  # ipywidgets.Output parameters

        return dc


    def __init__(self, experiment_data, experiment_descriptions=None, config=None, **kwargs):

        super().__init__(experiment_data,
                         experiment_descriptions=experiment_descriptions,
                         config=config,
                         **kwargs)

        # add a figure output below the selection
        self.figure_output = None

        #########################
        # handle config input if not a state backup was loaded
        if not hasattr(self, '_plot_function_key') and not hasattr(self, '_plot_function') and not hasattr(self, '_plot_function_configs'):

            # use first plot function in plot_fuctions dictionary if non is provided
            if self.config.plot_function == '' or self.config.plot_function is None:
                plot_function = list(self.config.plot_functions.keys())[0]
            else:
                plot_function = self.config.plot_function

            if isinstance(plot_function, str):
                # config.plot_function is a key for the config.plot_functions dictionary

                if self.config.plot_function not in self.config.plot_functions:
                    raise ValueError('If config.plot_function is a string then it must be a key for the config.plot_functions dictionary!')

                self._plot_function_key = plot_function
                self._plot_function = self.config.plot_functions[self._plot_function_key]

            elif callable(plot_function):
                # if the config.plotfunction is a function handle
                self._plot_function_key = 'NONE'
                self._plot_function = plot_function

            else:
                raise TypeError('Unsupported type for config.plot_fuction! Only strings or function handles are valid.')

            # handle self._plot_function_configs
            self._plot_function_configs = self.config.plot_function_configs.copy()
            plot_function_config = self.config.plot_function_config
            if plot_function_config == '' or plot_function_config is None:
                # if no config.plot_function_config is given, then use the one from the config.plot_function_configs dictionary
                # create a new entry in _plot_function_configs if non exists for this plot_function_key
                if self._plot_function_key not in self._plot_function_configs:
                    self._plot_function_configs[self._plot_function_key] = None
            else:
                # if a config is given, then override the _plot_functions_config with it
                self._plot_function_configs[self._plot_function_key] = _config_obj_to_dict(plot_function_config)

            # if the config is invalid and the user can not change it anymore, then raise an exception
            if not self.config.is_plot_function_config_editor \
                    and isinstance(self._plot_function_configs[self._plot_function_key], BaseException):
                raise self._plot_function_config[self._plot_function_key]

            if self._plot_function_key == 'NONE' and self.config.is_plot_function_selection:
                raise ValueError('If plot_function_selection is active, then the config.plot_function must be a key for the config.plot_functions dictionary!')

        ########################
        # add gui components
        selection_children = []

        # only allow selection of a plot function if the defined intial plot_function is not a function handle
        if self.config.is_plot_function_selection and callable(self.config.plot_function):
            self.config.is_plot_function_selection = False

        # selection of plot function
        if self.config.is_plot_function_selection:

            self.plot_function_selection_label_widget = ipywidgets.Label(**self.config.plot_function_selection.label)

            self.plot_function_selection_dropdown_widget = ipywidgets.Dropdown(
                options=list(self.config.plot_functions.keys()),
                value=self._plot_function_key,
                **self.config.plot_function_selection.dropdown)

            self.plot_function_selection_hbox_widget = ipywidgets.HBox(
                [self.plot_function_selection_label_widget, self.plot_function_selection_dropdown_widget],
                **self.config.plot_function_selection.hbox)

            selection_children.append(self.plot_function_selection_hbox_widget)

            # register event to know when a new plot function was selected
            self.plot_function_selection_dropdown_widget.observe(
                self._on_plot_function_selection_dropdown_widget_value_change,
                names='value')

        if self.config.is_plot_function_config_editor:

            self.plot_function_config_editor_textarea = ipywidgets.Textarea(
                value=self._plot_function_configs[self._plot_function_key],
                **self.config.plot_function_config_editor.textarea)

            self.plot_function_config_editor_accordion = ipywidgets.Accordion(
                children=[self.plot_function_config_editor_textarea],
                **self.config.plot_function_config_editor.accordion)
            self.plot_function_config_editor_accordion.set_title(0, self.config.plot_function_config_editor.title)

            selection_children.append(self.plot_function_config_editor_accordion)

            # register event to know when a new config was edited
            self.plot_function_config_editor_textarea.observe(
                self._on_plot_function_config_editor_value_change,
                names='value')

        # add selection items befor the box for the buttons
        eu.gui.jupyter.add_children_to_widget(self, selection_children, idx=-1)

        # add plotting button
        if self.config.is_plot_button:
            self.plot_button = ipywidgets.Button(**self.config.plot_button)
            self.plot_button.on_click(self._plot_button_on_click_handler)

            # append button to start of button box
            eu.gui.jupyter.add_children_to_widget(self.activity_hbox, self.plot_button, idx=0)


    def _on_plot_function_selection_dropdown_widget_value_change(self, event_descr):
        self._plot_function_key = event_descr['new']
        self._plot_function = self.config.plot_functions[self._plot_function_key]
        self.plot_function_config = self._plot_function_configs.get(self._plot_function_key, None)


    def _on_plot_function_config_editor_value_change(self, event_descr):
        self._plot_function_configs[self._plot_function_key] = event_descr['new']


    @property
    def plot_function(self):
        return self._plot_function


    @plot_function.setter
    def plot_function(self, plot_function):

        if isinstance(plot_function, str):
            # config.plot_function is a key for the config.plot_functions dictionary

            if plot_function not in self.config.plot_functions:
                raise ValueError(
                    'If plot_function is a string then it must be a key for the config.plot_functions dictionary!')

            self._plot_function_key = plot_function
            self._plot_function = self.config.plot_functions[self._plot_function_key]

            # also the config.plot_function_configs dict will be used
            self.plot_function_config = self._plot_function_configs[self._plot_function_key]

        elif callable(plot_function):
            # if the config.plotfunction is a function handle

            # try to find function in configured functions
            plot_function_key = 'NONE'
            for func_str, func in self.config.plot_functions.items():
                if func == plot_function:
                    plot_function_key = func_str
                    break

            self._plot_function_key = plot_function_key
            self._plot_function = plot_function
            if self._plot_function_key not in self._plot_function_configs:
                self._plot_function_configs[self._plot_function_key] = None

    @property
    def plot_function_config(self):
        return _config_obj_to_dict(self._plot_function_configs[self._plot_function_key])


    @plot_function_config.setter
    def plot_function_config(self, plot_function_config):

        if self.config.is_plot_function_config_editor:

            if not isinstance(plot_function_config, str):
                raise ValueError('If the plot_function_config_editor is activated, then the given plot_function_config must be a string!')

            # was the gui element already created?
            if hasattr(self, 'plot_function_config_editor_textarea'):
                self.plot_function_config_editor_textarea.value = plot_function_config

        self._plot_function_configs[self._plot_function_key] = plot_function_config


    def _plot_button_on_click_handler(self, _):
        self.plot_data()


    @property
    def selection(self):
        '''AttrDict (dict) with the selected experiment data options.'''
        selection = super().selection
        selection.plot_function = self.plot_function
        selection.plot_function_config = self.plot_function_config
        return selection


    @selection.setter
    def selection(self, selection):
        super(self.__class__, self.__class__).selection.fset(self, selection)
        if 'plot_function' in selection: self.plot_function = selection.plot_function
        if 'plot_function_config' in selection: self.plot_function_config = selection.plot_function_config


    def get_widget_state(self):
        state = super().get_widget_state()

        if self._plot_function_key == 'NONE':
            state.plot_function = self.plot_function
        else:
            state.plot_function = self._plot_function_key
        state._plot_function_configs = self._plot_function_configs
        return state


    def set_widget_state(self, state):

        if '_plot_function_configs' in state:
            self._plot_function_configs = state._plot_function_configs  # set all gui configs

        if 'plot_function' in state:
            self.plot_function = state.plot_function

        if '_plot_function_configs' in state:
            self.plot_function_config = self._plot_function_configs[self._plot_function_key]  # update the view

        return super().set_widget_state(state)


    def plot_data(self):

        # display the figure
        # must be created here so that the display is created below the selection gui
        if self.figure_output is None:
            self.figure_output = ipywidgets.Output(**self.config.figure_output)
            IPython.display.display(self.figure_output)
        else:
            self.figure_output.clear_output(wait=True)

        with self.figure_output:
            print('Plotting ...')

            # load experimental data before plotting
            self.select_experiment_data()

            plot_config = self.plot_function_config
            if isinstance(plot_config, BaseException):
                raise plot_config
            else:
                display_obiect = self.plot_function(
                    self.selected_data,
                    labels=self.selected_data_labels,
                    config=plot_config)

                self.figure_output.clear_output(wait=True)
                IPython.display.display(display_obiect)


    def get_code_producer_variables(self):

        variables = super().get_code_producer_variables()

        plot_function_handle = self.plot_function
        variables.plot_function = plot_function_handle.__name__

        if plot_function_handle.__module__ == '__main__' or plot_function_handle.__module__ == 'builtins':
            # no extra imports
            variables.import_statements = '\n'
        else:
            variables.import_statements = 'from {} import {}'.format(plot_function_handle.__module__, variables.plot_function)

        # plot configuration as a dictionary
        if self.config.is_plot_function_config_editor:
            plot_function_config_str = 'eu.AttrDict(\n{})'.format(self.plot_function_config_editor_textarea.value)
        else:
            plot_function_config_dict = self.plot_function_config
            plot_function_config_str = 'eu.' + str(eu.AttrDict(plot_function_config_dict))

        variables.plot_function_config = plot_function_config_str

        return variables


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import dill
import os
from glob import glob

DILL_FILE_EXTENSION = 'dill'

def save_dill(obj,
              file_path: str):
    """
    Serializes a Python object and saves it to a file using the [dill](https://pypi.org/project/dill/)
    serialization library.

    Parameters:
        obj (Any):
            The Python object to be serialized.
        file_path (str):
            The file path where the serialized object will be saved.

    <h4>Notes:</h4>

    - If the provided file path does not have the correct file extension for Dill files,
      the extension will be added automatically.
    - The necessary directories for the file path will be created if they do not exist.
    """
    if not file_path.endswith('.' + DILL_FILE_EXTENSION):
        file_path += '.' + DILL_FILE_EXTENSION

    eu.io.makedirs_for_file(file_path)
    with open(file_path, 'wb') as fh:
        dill.dump(obj, fh)


def load_dill(file_path: str) -> object:
    """
    Loads a serialized object from a file using the [dill](https://pypi.org/project/dill/) library.

    Parameters:
        file_path (str):
            The path to the file from which to load the object.
            The file extension is optionally added if not already present.

    Returns:
        obj (Any): The object that was deserialized from the file.

    <h4>Notes:</h4>

    - If the specified file does not exist, the function attempts to append
      the expected file extension (.dill) before throwing an error.
    - :warning: This could allow arbitrary code execution. Only load files you trust!
    """
    if not os.path.exists(file_path):
        if not file_path.endswith('.' + DILL_FILE_EXTENSION):
            file_path += '.' + DILL_FILE_EXTENSION

    with open(file_path, 'rb') as fh:
        obj = dill.load(fh)
    return obj


def load_dill_files(directory: str):
    """
    Loads all serialized objects from a directory using the [dill](https://pypi.org/project/dill/)
    library and returns them in a dictionary.

    Parameters:
        directory (str):
            The path to the directory containing dill-serialized files.

    Raises:
        FileNotFoundError: If the specified directory does not exist.

    Returns:
        data (AttrDict):
            An attribute dictionary where keys are the file names (without extensions) and
            values are the deserialized objects.

    <h4>Notes:</h4>

    - If the specified file does not exist, the function attempts to append
      the expected file extension (.dill) before throwing an error.
    - :warning: This could allow arbitrary code execution. Only load files you trust!
    """

    if not os.path.isdir(directory):
        raise FileNotFoundError('Directory {!r} does not exist!'.format(directory))

    data_dict = eu.AttrDict()

    for file in glob(os.path.join(directory, '*.' + DILL_FILE_EXTENSION)):
        data_name = os.path.splitext(os.path.basename(file))[0]
        data = load_dill(file)
        data_dict[data_name] = data

    return data_dict


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import glob
import os
import subprocess
import time
from typing import Optional, Union

import exputils
import numpy as np
from datetime import datetime
import fasteners

STATUS_FILE_EXTENSION = '.status'


def start_slurm_experiments(directory=None, start_scripts='*.slurm', is_parallel=True, verbose=False, post_start_wait_time=0):

    return start_experiments(
        directory=directory,
        start_scripts=start_scripts,
        start_command='sbatch {}',
        parallel=is_parallel,
        verbose=verbose,
        post_start_wait_time=post_start_wait_time,
        write_status_files_automatically=False
    )


def start_torque_experiments(directory=None, start_scripts='*.torque', is_parallel=True, verbose=False, post_start_wait_time=0):

    return start_experiments(
        directory=directory,
        start_scripts=start_scripts,
        start_command='qsub {}',
        parallel=is_parallel,
        verbose=verbose,
        post_start_wait_time=post_start_wait_time,
        write_status_files_automatically=False
    )


def start_experiments(directory: Optional[str] = None,
                      start_scripts: Optional[str] = 'run_*.py',
                      start_command: Optional[str] = '{}',
                      parallel: Union[bool, int] = True,
                      is_chdir: bool = True,
                      verbose: bool = False,
                      post_start_wait_time: float = 0.,
                      write_status_files_automatically: bool = True):
    """
    Searches all the start scripts of experiments and/or repetitions in the experiments folder
    and executes them either in parallel or sequentially.

    It also documents their execution status (todo, running, finished, error) in a status file
    that allows it to identify if a script should be executed or not when used again on the
    same target directory.

    Parameters:
        directory (str):
            Directory in which the start scripts are searched.
            Default is `'./experiments'`.
        start_scripts (str):
            Filename of the start script file that are searched under the given target directory.
            Can include '*' to search for scripts, for example 'run_*.py'.
            The default `'run_*'` will look for all files that start with 'run' and try to start them.
        parallel (bool, int):
            Defines if scripts should be started in parallel and how many are allowed to run in parallel.
            If `False` then the scripts are started sequentially one after another.
            If `True` then the scripts are started and executed in parallel all at once.
            If an integer, then the number defines how many scripts can run in parallel.
        is_chdir (bool):
            Before starting a script, should the main process change to its working directory.
        verbose (bool):
            Should verbose output with more information given. Default is `False`.
        post_start_wait_time (float):
            Time waited before one process is started after another.
        write_status_files_automatically (bool):
            Should status files that document if scripts were started and executed be
            written by the manager. These are important to identify if an experiment or repetition
            did run already.
    """

    if directory is None:
        directory = os.path.join('.', exputils.DEFAULT_EXPERIMENTS_DIRECTORY)

    # handle number of parallel processes
    if isinstance(parallel, bool):
        if parallel:
            n_parallel = np.inf
        else:
            n_parallel = 1
    elif isinstance(parallel, int):
        if parallel <= 0:
            raise ValueError('Number of parallel processes must be larger 0!')
        else:
            n_parallel = parallel
    else:
        raise ValueError('Argument \'parallel\' must be either a bool or an integer number!')

    if is_chdir:
        cwd = os.getcwd()

    # get all scripts
    all_scripts = get_scripts(directory=directory, start_scripts=start_scripts)

    ignored_scripts = []
    todo_scripts = []
    # check their initial status and write one for the scripts that will be started
    for script in all_scripts:

        # lock processing of the script, so that no other running experimentstarter is updating its status in parallel
        with _get_script_lock(script):

            status = get_script_status(script)

            if status is None:
                if write_status_files_automatically:
                    _update_script_status(script, 'todo')
                todo_scripts.append(script)

            elif status.lower() != 'finished':
                todo_scripts.append(script)

            else:
                ignored_scripts.append((script, status))

    # start all in parallel if wanted
    if n_parallel == np.inf:
        n_parallel = len(todo_scripts)

    # started process and their corresponding scripts
    started_processes = []
    started_scripts = []
    finished_processes_idxs = []

    next_todo_script_idx = 0
    n_active_processes = 0

    # run as long as there is an active process or we did not finish all processes yet
    while n_active_processes > 0 or next_todo_script_idx < len(todo_scripts):

        # start as many processes as parallel processes are allowed
        for i in range(n_parallel - n_active_processes):

            # stop starting processes when all scripts are started
            if next_todo_script_idx < len(todo_scripts):

                script = todo_scripts[next_todo_script_idx]
                next_todo_script_idx += 1

                # lock processing of the script, so that no other running experimentstarter is starting it in parallel
                with _get_script_lock(script):

                    # check the script status, only start if needed
                    status = get_script_status(script)
                    if _is_to_start_status(status):

                        if write_status_files_automatically:
                            _update_script_status(script, 'running')

                        # start
                        script_directory = os.path.dirname(script)
                        script_path_in_its_working_directory = os.path.join('.', os.path.basename(script))

                        print('{} start {!r} (previous status: {}) ...'.format(datetime.now().strftime("%Y/%m/%d %H:%M:%S"), script, status))

                        process_environ = {
                            **os.environ,
                            "EU_STATUS_FILE": script_path_in_its_working_directory + STATUS_FILE_EXTENSION,
                        }

                        if is_chdir:
                            os.chdir(script_directory)
                            process = subprocess.Popen(start_command.format(script_path_in_its_working_directory).split(), env=process_environ)
                            os.chdir(cwd)
                        else:
                            process = subprocess.Popen(start_command.format(script).split(), cwd=script_directory, env=process_environ)

                        started_processes.append(process)
                        started_scripts.append(script)

                        if post_start_wait_time > 0:
                            time.sleep(post_start_wait_time)

                    else:
                        # do not start
                        ignored_scripts.append((script, status))

        # check the activity of the started processes
        n_active_processes = 0
        for p_idx, process in enumerate(started_processes):

            if p_idx not in finished_processes_idxs:

                if process.poll() is None:
                    n_active_processes += 1
                else:
                    finished_processes_idxs.append(p_idx)
                    if process.returncode == 0:
                        status = 'finished'
                    else:
                        status = 'error'

                    if write_status_files_automatically:
                        _update_script_status(started_scripts[p_idx], status)

                    print('{} finished {!r} (status: {})'.format(datetime.now().strftime("%Y/%m/%d %H:%M:%S"), started_scripts[p_idx], status))

        if n_active_processes > 0:
            time.sleep(0.5) # sleep half a second before checking again

    if verbose:
        if ignored_scripts:
            print('Ignored scripts:')
            for (script_path, status) in ignored_scripts:
                print('\t- {!r} (status: {})'.format(script_path, status))


def _is_to_start_status(status):
    """Returns true if the given status means that the script should be started, otherwise false."""
    return status is None or status.lower().startswith('todo') or status.lower().startswith('none') or status.lower().startswith('error') or status.lower().startswith('unfinished')


def _get_script_lock(script):
    """Create a lock for the given script that can be used to have exclusive access to write its status."""
    return fasteners.InterProcessLock(script + '.lock')


def _update_script_status(script, status):
    """
    Updates the status for the given script.

     :param script: Path to the script.
     :param status: New status.
    """

    status_file_path = script + STATUS_FILE_EXTENSION

    time_str = datetime.now().strftime("%Y/%m/%d %H:%M:%S")

    with open(status_file_path, 'a+') as file:
        file.write( time_str + "\n" + status + "\n")


def get_scripts(directory: Optional[str] = None,
                start_scripts: Optional[str] = 'run_*.py') -> list:
    """
    Searches all start scripts in the experiments directory.

    Parameters:
        directory (str):
            Directory in which the start scripts are searched.
            Default is `'./experiments'`.
        start_scripts (str):
            Filename of the start script file that are searched under the given target directory.
            Can include '*' to search for scripts, for example 'run_*.py'.
            The default `'run_*'` will look for all files that start with 'run' and try to start them.

    Returns:
        scripts (list): List of filepaths to the start scripts.
    """

    if directory is None:
        directory = os.path.join('.', exputils.DEFAULT_EXPERIMENTS_DIRECTORY)

    # find all start scripts
    scripts = glob.iglob(os.path.join(directory, '**', start_scripts), recursive=True)
    scripts = list(scripts)
    scripts.sort()

    return scripts


def get_script_status(script_file: str) -> Optional[str]:
    """
    Returns the execution status of a certain start script.

    Parameters:
        script_file (str): Path to the script file.

    Returns:
        status (str, None):
            Status as a string. Usually `'todo'`, `'error'`, `'running'`, or `'finished'`.
            `None` if no status exists.
    """
    status = None

    status_file_path = script_file + STATUS_FILE_EXTENSION

    if os.path.isfile(status_file_path):
        # read status
        with open(status_file_path, 'r') as f:
            lines = f.read().splitlines()
            if len(lines) > 0:
                status = lines[-1]

    return status


def get_number_of_scripts_to_execute(directory: Optional[str] = None,
                                     start_scripts: str = 'run_*.py') -> int:
    """
    Identifies the number of scripts that have to be executed in the experiments directory.
    Scripts that have to be executed have either the status 'none', 'todo', 'error', or 'unfinished'.

    Parameters:
        directory (str):
            Directory in which the start scripts are searched.
            Default is `'./experiments'`.
        start_scripts (str):
            Filename of the start script file that are searched under the given target directory.
            Can include '*' to search for scripts, for example 'run_*.py'.
            The default `'run_*'` will look for all files that start with 'run' and try to start them.

    Returns:
        n_scripts (int): Number of scripts that have to be executed.
    """

    scripts = get_scripts(directory=directory, start_scripts=start_scripts)

    n = 0
    for script in scripts:
        status = get_script_status(script)
        if _is_to_start_status(status):
            n += 1

    return n


def get_number_of_scripts(directory: Optional[str] = None,
                          start_scripts: str = 'run_*.py'):
    """
    Identifies the number of all scripts in the experiments directory regardless of their execution
    status.

    Parameters:
        directory (str):
            Directory in which the start scripts are searched.
            Default is `'./experiments'`.
        start_scripts (str):
            Filename of the start script file that are searched under the given target directory.
            Can include '*' to search for scripts, for example 'run_*.py'.
            The default `'run_*'` will look for all files that start with 'run' and try to start them.

    Returns:
        n_scripts (int): Number of scripts.
    """

    scripts = get_scripts(directory=directory, start_scripts=start_scripts)
    return len(scripts)







##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
from exputils.gui.jupyter.experiment_data_loader_widget import ExperimentDataLoaderWidget
from exputils.gui.jupyter.multi_selection_widget import MultiSelectionWidget
from exputils.gui.jupyter.experiment_ids_selection_widget import ExperimentIDsSelectionWidget
from exputils.gui.jupyter.repetition_ids_selection_widget import RepetitionIDsSelectionWidget
from exputils.gui.jupyter.experiment_data_selection_widget import ExperimentDataSelectionWidget
from exputils.gui.jupyter.experiment_data_plot_selection_widget import ExperimentDataPlotSelectionWidget
from exputils.gui.jupyter.text_selection_accordion_widget import TextSelectionAccordionWidget
from exputils.gui.jupyter.base_widget import BaseWidget

from exputils.gui.jupyter.plotly_meanstd_scatter import plotly_meanstd_scatter
from exputils.gui.jupyter.plotly_meanstd_bar import plotly_meanstd_bar
from exputils.gui.jupyter.plotly_box import plotly_box
from exputils.gui.jupyter.tabulate_meanstd import tabulate_meanstd
from exputils.gui.jupyter.tabulate_pairwise import tabulate_pairwise

from exputils.gui.jupyter.misc import create_new_cell
from exputils.gui.jupyter.misc import save_config
from exputils.gui.jupyter.misc import load_config
from exputils.gui.jupyter.misc import add_children_to_widget
from exputils.gui.jupyter.misc import remove_children_from_widget
from exputils.gui.jupyter.misc import set_children_of_widget
from exputils.gui.jupyter.misc import generate_random_state_backup_name

DEFAULT_CONFIG_DIRECTORY = '.ipython_config'

# TODO: Refactor - create plotly plots as objects
# TODO: Feature - create table widget that shows statistical significant difference measures for same data that is given to bar or box plots





##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils.gui.jupyter
import exputils.gui.misc


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import ipywidgets
import warnings
from exputils.gui.jupyter.base_widget import BaseWidget
from exputils.gui.jupyter.experiment_ids_selection_widget import ExperimentIDsSelectionWidget
from exputils.gui.jupyter.repetition_ids_selection_widget import RepetitionIDsSelectionWidget
from exputils.gui.jupyter.code_producer_widget import CodeProducerWidget

# TODO: Bug - giving several datasources seems not not work, as they are combined to a single string
# TODO: Feature - add selection of label_templates for data sources, experiments, and repetitions
# TODO: Feature - add data source selection helper
# TODO: Feature - allow that changes of names in the experiment_data_loader are directly taken as updates here
# TODO: Refactor - put the output widget into this class and allow to print errors and warnings to it if data is loaded
# TODO: Feature - add on_selection_changed event
# TODO: Refactor - keep for each selection attribute (e.g. datasources) an internal variable (self._datasources) which
#       is updated if the selection changes. This allows to not overrider the config in the case of default values.

CODE_TEMPLATE_MULTILINE = """selection_widget = ExperimentDataSelectionWidget(
    experiment_data, 
    experiment_descriptions,
    datasources=<datasources>,
    experiment_ids=<experiment_ids>,
    repetition_ids=<repetition_ids>,
    output_format=<output_format>,
    data_filter=<data_filter>,
    state_backup_name=<state_backup_name>,
    state_backup_variable_filter=['experiment_ids', 'repetition_ids'],
    is_datasources_selection=False,
    is_output_format_selection=False,
    is_data_filter_selection=False,
    is_code_producer=False) 
display(selection_widget)"""

CODE_TEMPLATE_SINGLELINE = """selection_widget = ExperimentDataSelectionWidget(experiment_data, experiment_descriptions, datasources=<datasources>, experiment_ids=<experiment_ids>, repetition_ids=<repetition_ids>, output_format=<output_format>, data_filter=<data_filter>, state_backup_name=<state_backup_name>, is_datasources_selection=False, is_output_format_selection=False, is_data_filter_selection=False, is_code_producer=False)
display(selection_widget)"""


class ExperimentDataSelectionWidget(BaseWidget, ipywidgets.VBox):

    @staticmethod
    def default_config():
        dc = BaseWidget.default_config()

        dc.is_datasources_selection = True
        dc.datasources = ''
        dc.datasources_selection = eu.AttrDict(
            hbox=eu.AttrDict(layout=eu.AttrDict(width='100%')),  # ipywidgets.HBox parameters
            label=eu.AttrDict(  # ipywidgets.Label parameters
                value='Data Sources:',
                layout=eu.AttrDict(min_width='100px')),
            text=eu.AttrDict(  # ipywidgets.Text parameters
                placeholder='Provide a comma seperated list of data sources ...',
                description='',
                layout=eu.AttrDict(width='100%')))

        dc.is_experiment_ids_selection = True
        dc.experiment_ids = 'all'
        dc.experiment_ids_selection = eu.AttrDict(
            hbox=eu.AttrDict(layout=eu.AttrDict(width='100%')),  # ipywidgets.HBox parameters
            label=eu.AttrDict(  # ipywidgets.Label parameters
                value='Experiments:',
                layout=eu.AttrDict(min_width='100px')),
            multiselect=eu.AttrDict(  # exputils.gui.jupyter.ExperimentIDsSelectionWidget parameters
                main_vbox=eu.AttrDict(
                    layout=eu.AttrDict(width='100%'))))

        dc.is_repetition_ids_selection = True
        dc.repetition_ids = 'all'
        dc.repetition_ids_selection = eu.AttrDict(
            hbox=eu.AttrDict(layout=eu.AttrDict(width='100%')),  # ipywidgets.HBox parameters
            label=eu.AttrDict(  # ipywidgets.Label parameters
                value='Repetitions:',
                layout=eu.AttrDict(min_width='100px')),
            multiselect=eu.AttrDict(  # exputils.gui.jupyter.MultiSelectionWidget parameters
                main_vbox=eu.AttrDict(
                    layout=eu.AttrDict(width='100%'))))

        dc.is_output_format_selection = False
        dc.output_format = 'S,E,D'
        dc.output_format_selection = eu.AttrDict(
            hbox=eu.AttrDict(layout=eu.AttrDict(width='100%')),  # ipywidgets.HBox parameters
            label=eu.AttrDict(  # ipywidgets.Label parameters
                value='Output Format:',
                layout=eu.AttrDict(min_width='100px')),
            text=eu.AttrDict(  # ipywidgets.Text parameters
                placeholder='Provide output format description string ...',
                description='',
                layout=eu.AttrDict(width='100%')))

        dc.is_data_filter_selection = False
        dc.data_filter = ''
        dc.data_filter_selection = eu.AttrDict(
            hbox=eu.AttrDict(layout=eu.AttrDict(width='100%')),  # ipywidgets.HBox parameters
            label=eu.AttrDict(  # ipywidgets.Label parameters
                value='Filter:',
                layout=eu.AttrDict(min_width='100px')),
            text=eu.AttrDict(  # ipywidgets.Text parameters
                placeholder='Provide optional filter ...',
                description='',
                layout=eu.AttrDict(width='100%')))
        dc.main_vbox = eu.AttrDict()

        # Box that contains the elements that do activities
        dc.activity_hbox = eu.AttrDict(
            layout=eu.AttrDict(width='100%', height='100%')
        )

        dc.is_get_experiment_data_button = True
        dc.get_experiment_data_button = eu.AttrDict(
            description='Get Data',
            tooltip='Collects the exoerimental data according to the given selection.',
            layout=eu.AttrDict(width='100%', height='95%'))

        dc.is_code_producer = True
        dc.code_producer = eu.AttrDict(
            code_templates=[dict(name='Multi Line', code_template=CODE_TEMPLATE_MULTILINE),
                            dict(name='Single Line', code_template=CODE_TEMPLATE_SINGLELINE)],
            main_box=eu.AttrDict(layout=eu.AttrDict(width='25%', height='99%')))

        # config for collecting the data
        dc.select_experiment_data_func = eu.data.select_experiment_data  # function to get the data

        return dc


    def __init__(self, experiment_data_object, experiment_descriptions=None, config=None, **kwargs):

        # constructor of BaseWidget
        super().__init__(config=config, **kwargs)
        # constructor of VBox
        super(BaseWidget, self).__init__(
            **self.config.main_vbox)

        # handle input of either data loader widget or data itsel
        if isinstance(experiment_data_object, eu.gui.jupyter.ExperimentDataLoaderWidget):
            if experiment_descriptions is not None:
                warnings.warn('If an ExperimentDataLoaderWidget is provided, then the experiment_descriptions parameter will be ignored.')

            self._experiment_data = experiment_data_object.experiment_data
            self._experiment_descriptions = experiment_data_object.experiment_descriptions

            # register event at widget, if it loads new data
            experiment_data_object.on_experiment_data_loaded(self._on_new_experiment_data_handler)
        else:
            self._experiment_data = experiment_data_object
            self._experiment_descriptions = experiment_descriptions

        self._selected_data = None
        self._selected_data_labels = None

        # self._on_selection_changed_event_handlers = []
        self._on_data_collected_event_handlers = []

        child_widgets = []

        if self.config.is_datasources_selection:
            self.datasources_selection_label_widget = ipywidgets.Label(**self.config.datasources_selection.label)
            self.datasources_selection_text_widget = ipywidgets.Text(
                value='',
                **self.config.datasources_selection.text)
            self.datasources_selection_hbox_widget = ipywidgets.HBox(
                [self.datasources_selection_label_widget, self.datasources_selection_text_widget],
                **self.config.datasources_selection.hbox)
            child_widgets.append(self.datasources_selection_hbox_widget)

        if self.config.is_experiment_ids_selection:
            self.experiment_ids_selection_label_widget = ipywidgets.Label(**self.config.experiment_ids_selection.label)
            self.experiment_ids_selection_multiselect_widget = ExperimentIDsSelectionWidget(
                self.experiment_data,
                experiment_descriptions=self.experiment_descriptions,
                **self.config.experiment_ids_selection.multiselect)
            self.experiment_ids_selection_hbox_widget = ipywidgets.HBox(
                [self.experiment_ids_selection_label_widget, self.experiment_ids_selection_multiselect_widget],
                **self.config.experiment_ids_selection.hbox)
            child_widgets.append(self.experiment_ids_selection_hbox_widget)

        if self.config.is_repetition_ids_selection:
            self.repetition_ids_selection_label_widget = ipywidgets.Label(**self.config.repetition_ids_selection.label)
            self.repetition_ids_selection_multiselect_widget = RepetitionIDsSelectionWidget(
                self.experiment_data,
                experiment_descriptions=self.experiment_descriptions,
                **self.config.repetition_ids_selection.multiselect)
            self.repetition_ids_selection_hbox_widget = ipywidgets.HBox(
                [self.repetition_ids_selection_label_widget, self.repetition_ids_selection_multiselect_widget],
                **self.config.repetition_ids_selection.hbox)
            child_widgets.append(self.repetition_ids_selection_hbox_widget)

        if self.config.is_output_format_selection:
            self.output_format_selection_label_widget = ipywidgets.Label(**self.config.output_format_selection.label)
            self.output_format_selection_text_widget = ipywidgets.Text(
                value='',
                **self.config.output_format_selection.text)
            self.output_format_selection_hbox_widget = ipywidgets.HBox(
                [self.output_format_selection_label_widget, self.output_format_selection_text_widget],
                **self.config.output_format_selection.hbox)
            child_widgets.append(self.output_format_selection_hbox_widget)

        if self.config.is_data_filter_selection:
            self.data_filter_selection_label_widget = ipywidgets.Label(**self.config.data_filter_selection.label)
            self.data_filter_selection_text_widget = ipywidgets.Text(
                value='',
                **self.config.data_filter_selection.text)
            self.data_filter_selection_hbox_widget = ipywidgets.HBox(
                [self.data_filter_selection_label_widget, self.data_filter_selection_text_widget],
                **self.config.data_filter_selection.hbox)
            child_widgets.append(self.data_filter_selection_hbox_widget)

        activity_box_child_widgets = []
        if self.config.is_get_experiment_data_button:
            self.get_experiment_data_button = ipywidgets.Button(**self.config.get_experiment_data_button)
            self.get_experiment_data_button.on_click(self._get_experiment_data_button_on_click_handler)
            activity_box_child_widgets.append(self.get_experiment_data_button)

        if self.config.is_code_producer:
            self.code_producer_widget = CodeProducerWidget(
                code_variables=self.get_code_producer_variables,
                **self.config.code_producer)
            activity_box_child_widgets.append(self.code_producer_widget)

        self.activity_hbox = ipywidgets.HBox(
            children=activity_box_child_widgets,
            **self.config.activity_hbox)
        child_widgets.append(self.activity_hbox)

        eu.gui.jupyter.add_children_to_widget(self, child_widgets)

        # set initial selection from config
        self.datasources = self.config.datasources
        self.experiment_ids = self.config.experiment_ids
        self.repetition_ids = self.config.repetition_ids
        self.output_format = self.config.output_format
        self.data_filter = self.config.data_filter

        # load previously backuped selection if exists
        self.load_state_backup()


    @property
    def experiment_data(self):
        return self._experiment_data


    @property
    def experiment_descriptions(self):
        return self._experiment_descriptions


    def set_experiment_data(self, experiment_data, experiment_descriptions=None):
        self._experiment_data = experiment_data
        self._experiment_descriptions = experiment_descriptions
        self._update_selections_to_new_experiment_data()


    @property
    def selected_data(self):
        return self._selected_data


    @property
    def selected_data_labels(self):
        return self._selected_data_labels


    @property
    def datasources(self):
        if not self.config.is_datasources_selection:
            return eu.misc.str_to_list(self.config.datasources)
        else:
            return eu.misc.str_to_list(self.datasources_selection_text_widget.value)


    @datasources.setter
    def datasources(self, datasources):

        if isinstance(datasources, str):
            datasources = eu.misc.str_to_list(datasources)

        if not self.config.is_datasources_selection:
            self.config.datasources = datasources
        else:
            self.datasources_selection_text_widget.value = ', '.join(datasources)


    @property
    def experiment_ids(self):
        if not self.config.is_experiment_ids_selection:
            return self.config.experiment_ids
        else:
            return self.experiment_ids_selection_multiselect_widget.selected_experiment_ids


    @experiment_ids.setter
    def experiment_ids(self, experiment_ids):
        if not self.config.is_experiment_ids_selection:
            self.config.experiment_ids = experiment_ids
        else:
            self.experiment_ids_selection_multiselect_widget.selected_experiment_ids = experiment_ids


    @property
    def repetition_ids(self):
        if not self.config.is_repetition_ids_selection:
            return self.config.repetition_ids
        else:
            rep_ids = self.repetition_ids_selection_multiselect_widget.selected_repetition_ids
            if not rep_ids:
                rep_ids = 'none'
            return rep_ids


    @repetition_ids.setter
    def repetition_ids(self, repetition_ids):
        if not self.config.is_repetition_ids_selection:
            self.config.repetition_ids = repetition_ids
        else:
            self.repetition_ids_selection_multiselect_widget.selected_repetition_ids = repetition_ids


    @property
    def output_format(self):

        if not self.config.is_output_format_selection:
            output_format = self.config.output_format
        else:
            output_format = self.output_format_selection_text_widget.value

        if isinstance(output_format, str):
            output_format = tuple(output_format.split(','))

        # transform string to tuple: 'S,E,D' --> ('S','E','D')
        return output_format


    @output_format.setter
    def output_format(self, output_format):

        if isinstance(output_format, str):
            output_format = tuple(output_format.replace(' ', '').split(','))
        elif isinstance(output_format, list):
            output_format = tuple(output_format)

        if not self.config.is_output_format_selection:
            self.config.output_format = output_format
        else:
            self.output_format_selection_text_widget.value = ', '.join(output_format)


    @property
    def data_filter(self):

        if not self.config.is_data_filter_selection:
            return self.config.data_filter
        else:
            return self.data_filter_text_widget.value


    @data_filter.setter
    def data_filter(self, data_filter):

        if not self.config.is_data_filter_selection:
            self.config.data_filter = data_filter
        else:
            self.data_filter_text_widget.value = data_filter


    @property
    def selection(self):
        '''AttrDict (dict) with the selected experiment data options.'''
        return eu.AttrDict(
            datasources=self.datasources,
            experiment_ids=self.experiment_ids,
            repetition_ids=self.repetition_ids,
            output_format=self.output_format,
            data_filter=self.data_filter)


    @selection.setter
    def selection(self, selection):
        selection = eu.AttrDict(selection)
        if 'datasources' in selection: self.datasources = selection.datasources
        if 'experiment_ids' in selection: self.experiment_ids = selection.experiment_ids
        if 'repetition_ids' in selection: self.repetition_ids = selection.repetition_ids
        if 'output_format' in selection: self.output_format = selection.output_format
        if 'data_filter' in selection: self.data_filter = selection.data_filter


    def on_data_collected(self, handler):
        '''
        Register an event handler for the case new data was collected.
        The handler receives a dict with information about the event.
        '''
        self._on_data_collected_event_handlers.append(handler)


    def _call_data_collected_event(self):
        descr = {'name': 'data_collected',
                 'owner': self,
                 'type': 'change'}

        for handler in self._on_data_collected_event_handlers:
            handler(descr)


    def select_experiment_data(self):

        data = self.config.select_experiment_data_func(
            self.experiment_data,
            self.datasources,
            experiment_ids=self.experiment_ids,
            repetition_ids=self.repetition_ids,
            output_format=self.output_format,
            experiment_descriptions=self.experiment_descriptions)

        # some get data functions provide labels, some not
        if isinstance(data, tuple):
            self._selected_data = data[0]
            self._selected_data_labels = data[1]
        else:
            self._selected_data = data
            self._selected_data_labels = None

        self._call_data_collected_event()

        self.backup_state()

        return self.selected_data, self.selected_data_labels


    def _get_experiment_data_button_on_click_handler(self, _):
        self.select_experiment_data()


    def _code_production_button_on_click_handler(self, _):
        eu.gui.jupyter.create_new_cell(r"print('hello world')")


    def get_widget_state(self):
        state = super().get_widget_state()
        state.datasources = self.datasources
        state.experiment_ids = self.experiment_ids
        state.repetition_ids = self.repetition_ids
        state.output_format = self.output_format
        state.data_filter = self.data_filter
        return state


    def set_widget_state(self, state):
        if 'datasources' in state: self.datasources = state.datasources
        if 'experiment_ids' in state: self.experiment_ids = state.experiment_ids
        if 'repetition_ids' in state: self.repetition_ids = state.repetition_ids
        if 'output_format' in state: self.output_format = state.output_format
        if 'data_filter' in state: self.data_filter = state.data_filter
        return super().set_widget_state(state)


    def get_code_producer_variables(self):

        self.backup_state()

        if isinstance(self.experiment_ids, list):
            experiment_ids = str(self.experiment_ids)
        else:
            experiment_ids = r"'{}'".format(self.experiment_ids)

        if isinstance(self.repetition_ids, list):
            repetition_ids = str(self.repetition_ids)
        else:
            repetition_ids = r"'{}'".format(self.repetition_ids)

        return eu.AttrDict(
            datasources=str(self.datasources),
            experiment_ids=experiment_ids,
            repetition_ids=repetition_ids,
            output_format=self.output_format,
            data_filter=r"'{}'".format(self.data_filter),
            state_backup_name=r"'{}'".format(eu.gui.jupyter.misc.generate_random_state_backup_name()))


    def _on_new_experiment_data_handler(self, event_descr):
        '''
        Called if the experiment data loader widget loaded new experiment data.
        Updates the experiment_data of the selection widget.
        '''
        self.set_experiment_data(
            event_descr['owner'].experiment_data,
            event_descr['owner'].experiment_descriptions)


    def _update_selections_to_new_experiment_data(self):

        # create new experiment_id_selection
        if self.config.is_experiment_ids_selection:
            self.experiment_ids_selection_multiselect_widget = ExperimentIDsSelectionWidget(
                self.experiment_data,
                experiment_descriptions=self.experiment_descriptions,
                **self.config.experiment_ids_selection.multiselect)
            eu.gui.jupyter.set_children_of_widget(
                self.experiment_ids_selection_hbox_widget,
                1,
                self.experiment_ids_selection_multiselect_widget)

        # create new repetition_id_selection
        if self.config.is_repetition_ids_selection:
            self.repetition_ids_selection_multiselect_widget = RepetitionIDsSelectionWidget(
                self.experiment_data,
                experiment_descriptions=self.experiment_descriptions,
                **self.config.repetition_ids_selection.multiselect)
            eu.gui.jupyter.set_children_of_widget(
                self.repetition_ids_selection_hbox_widget,
                1,
                self.repetition_ids_selection_multiselect_widget)


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
# Code from https://github.com/marcoconti83/read-ods-with-odfpy
# FILE: ODSReader.py

# Copyright 2011 Marco Conti

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTICE OF CHANGES:
#  - removed getSheet method
#  - renamed ODSReader.SHEETS to ODSReader.sheets
#  - read out c.data instead of n.data in line 82

# Thanks to grt for the fixes

import odf.opendocument
from odf.table import Table, TableRow, TableCell
from odf.text import P


# http://stackoverflow.com/a/4544699/1846474
class GrowingList(list):
    def __setitem__(self, index, value):
        if index >= len(self):
            self.extend([None]*(index + 1 - len(self)))
        list.__setitem__(self, index, value)


class ODSReader:

    # loads the file
    def __init__(self, file, clonespannedcolumns=None):
        self.clonespannedcolumns = clonespannedcolumns
        self.doc = odf.opendocument.load(file)
        self.sheets = {}
        for sheet in self.doc.spreadsheet.getElementsByType(Table):
            self.readSheet(sheet)

    # reads a sheet in the sheet dictionary, storing each sheet as an
    # array (rows) of arrays (columns)
    def readSheet(self, sheet):
        name = sheet.getAttribute("name")
        rows = sheet.getElementsByType(TableRow)
        arrRows = []

        # for each row
        for row in rows:
            row_comment = ""
            arrCells = GrowingList()
            cells = row.getElementsByType(TableCell)

            # for each cell
            count = 0
            for cell in cells:
                # repeated value?
                repeat = cell.getAttribute("numbercolumnsrepeated")
                if(not repeat):
                    repeat = 1
                    spanned = int(cell.getAttribute('numbercolumnsspanned') or 0)
                    # clone spanned cells
                    if self.clonespannedcolumns is not None and spanned > 1:
                        repeat = spanned

                ps = cell.getElementsByType(P)
                textContent = ""

                # for each text/text:span node
                for p in ps:
                    for n in p.childNodes:
                        if (n.nodeType == 1 and n.tagName == "text:span"):
                            for c in n.childNodes:
                                if (c.nodeType == 3):
                                    textContent = u'{}{}'.format(textContent, c.data)

                        if (n.nodeType == 3):
                            textContent = u'{}{}'.format(textContent, n.data)

                if(textContent):
                    if(textContent[0] != "#"):  # ignore comments cells
                        for rr in range(int(repeat)):  # repeated?
                            arrCells[count]=textContent
                            count+=1
                    else:
                        row_comment = row_comment + textContent + " "
                else:
                    for rr in range(int(repeat)):
                        count+=1

            # if row contained something
            if(len(arrCells)):
                arrRows.append(arrCells)

            #else:
            #    print ("Empty or commented row (", row_comment, ")")

        self.sheets[name] = arrRows


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import numpy as np
import os
import glob
import zipfile
import exputils
import re

def calc_repetition_statistics(statistics, load_experiment_data_func,  *args, statistics_directory=None,
                               recalculate_statistics=False, verbose=False, repetition_directory_template=None,
                               experiment_directory_template=None):
    '''
    Calculates statistics of a single repetition of an experiment.

    :param statistics: List with tuples of the form: (statistic name, statistic function)
    :param args: Directoryies in which the experiments are for which the statistics should be computed.
    :param results_directory:
    :param statistics_directory:
    :return:
    '''

    if len(args) == 0:
        experiments = ['.']
    elif len(args) == 1 and isinstance(args[0], list):
        experiments = args[0]
    else:
        experiments = list(args)

    if repetition_directory_template is None: repetition_directory_template = exputils.REPETITION_DIRECTORY_TEMPLATE
    repetition_directory_template = re.sub('\{.*\}', '', repetition_directory_template)

    if experiment_directory_template is None: experiment_directory_template = exputils.EXPERIMENT_DIRECTORY_TEMPLATE
    experiment_directory_template = re.sub('\{.*\}', '', experiment_directory_template)

    if statistics_directory is None: statistics_directory = exputils.DEFAULT_DATA_DIRECTORY

    def get_data(data, experiment_folder):
        '''Loads the data if the data is not already loaded'''
        if data is None:
            data = load_experiment_data_func(experiment_folder)
        return data

    # identify the experiment folders
    experiment_folders = []
    for folder in experiments:

        found_folders = []

        # make sure the folder has folder form, i.e. with final '/'
        # then take the base directory and search in it
        # this makes sure, that also the direct given folder is looked after, e.g. './experiment_000001' would also be identified
        basedir = os.path.dirname(os.path.join(folder, ''))
        directory_name = os.path.basename(os.path.dirname(basedir))

        # look if directory itself is a repetition experiment
        if directory_name.find(repetition_directory_template) >=0 and os.path.isdir(basedir):
            found_folders.append(basedir)

        # look if there are repetition folders somewhere inside the directory
        if not found_folders:
            for foldername in glob.iglob(os.path.join(basedir,'**',repetition_directory_template + '*'), recursive=True):
                if os.path.isdir(foldername):
                    found_folders.append(foldername)

        # if there are no repetitions, then search for experiment folder
        if not found_folders:
            if directory_name.find('experiment_') >= 0 and os.path.isdir(basedir):
                found_folders.append(basedir)

        if not found_folders:
            for foldername in glob.iglob(os.path.join(basedir, '**', experiment_directory_template + '*'), recursive=True):
                if os.path.isdir(foldername):
                    found_folders.append(foldername)

        if not found_folders:
            # assume the given folder is the the one
            found_folders.append(folder)

        experiment_folders.extend(found_folders)

    # calc statistic if it does not exist already
    for experiment_folder in sorted(experiment_folders):

        data = None

        directory = os.path.join(experiment_folder, statistics_directory)

        if not os.path.isdir(directory):
            os.makedirs(directory)

        if verbose:
            print('Calculate statistics for {!r}:'.format(experiment_folder))

        for statistic_definition in statistics:

            if isinstance(statistic_definition, tuple):
                statistic_name = statistic_definition[0]
                statistic_func = statistic_definition[1]
                statistic_type = statistic_definition[2] if len(statistic_definition) > 2 else 'numpy'
            elif isinstance(statistic_definition, dict):
                statistic_name = statistic_definition['name']
                statistic_func = statistic_definition['function']
                statistic_type = statistic_definition['type'] if 'type' in statistic_definition else 'numpy'
            else:
                raise ValueError('Unknown format for statistic definition {!r}!'.format(statistic_definition))

            # calculate statistics if they do not exist
            filepath_npy = os.path.join(directory, '{}.npy'.format(statistic_name))
            filepath_npz = os.path.join(directory, '{}.npz'.format(statistic_name))
            filepath_zip = os.path.join(directory, '{}.zip'.format(statistic_name))
            directory_path = os.path.join(directory, statistic_name)

            if (not os.path.isfile(filepath_npy) and not os.path.isfile(filepath_npz) and not os.path.isfile(filepath_zip) and not os.path.isdir(directory_path)) or recalculate_statistics:

                if verbose:
                    print('\t{} ...'.format(statistic_name))

                data = get_data(data, experiment_folder)

                if statistic_type == 'numpy':
                    stat = statistic_func(statistic_name, data)

                    if isinstance(stat, dict):
                        np.savez(filepath_npz, **stat)
                    else:
                        np.save(filepath_npy, stat)

                elif statistic_type == 'zip':

                    stat = statistic_func(statistic_name, data)

                    if not isinstance(stat, dict):
                        raise ValueError('Only dictionaries are accepted as data type for zip statistics!')

                    zf = zipfile.ZipFile(filepath_zip,
                                         mode='w',
                                         compression=zipfile.ZIP_DEFLATED,
                                         )
                    try:
                        for sub_stat_name, sub_stat in stat.items():
                            zf.writestr(sub_stat_name, sub_stat)

                    finally:
                        zf.close()

                elif statistic_type == 'directory':
                    if not os.path.isdir(directory_path):
                        os.mkdir(directory_path)
                    statistic_func(statistic_name, data, directory_path)

                else:
                    ValueError('Unknown statistic type {!r}!'.format(statistic_type))



def calc_statistics_over_repetitions(statistics, *args, load_data_func=None, recalculate_statistics=False, verbose=False, repetition_directory_template=None, statistics_directory=None):
    '''
    Calculates the statistics over several repetitions of an experiments.

    :param statistics: List with tuples of the form: (statistic name, statistic function)
    :param args: Directories in which the experiments are for which the statistics should be computed.
    :param results_directory:
    :param statistics_directory:
    :return:
    '''

    if len(args) == 0:
        experiments = ['.']
    elif len(args) == 1 and isinstance(args[0], list):
        experiments = args[0]
    else:
        experiments = list(args)

    if load_data_func is None: load_data_func = load_data_from_repetitions

    if repetition_directory_template is None: repetition_directory_template = exputils.REPETITION_DIRECTORY_TEMPLATE
    repetition_directory_template = re.sub('\{.*\}', '*', repetition_directory_template)

    if statistics_directory is None: statistics_directory = exputils.DEFAULT_DATA_DIRECTORY

    def get_data(data, experiment_folder):
        '''Loads the data if the data is not already loaded'''
        if data is None:
            data = load_data_func(experiment_folder)
        return data

    # identify the experiment folders, i.e. folders that have repetition folders as subfolders
    experiment_directories = dict()
    for folder in experiments:

        # make sure the folder has folder form, i.e. with final '/'
        # then take the base directory and search in it
        # this makes sure, that also the direct given folder is looked after, e.g. './experiment_000001' would also be identified
        basedir = os.path.dirname(os.path.join(folder, ''))
        
        # look if there are repetition folders somewhere inside the directory
        for repetition_directory in glob.iglob(os.path.join(basedir,'**', repetition_directory_template), recursive=True):
            if os.path.isdir(repetition_directory):
                # remember the parent folder, i.e. the experiment folder
                experiment_directory = os.path.dirname(repetition_directory)
                
                if experiment_directory not in experiment_directories:
                    experiment_directories[experiment_directory] = set()

                experiment_directories[experiment_directory].add(repetition_directory)

    # calc statistic if it does not exist already
    for experiment_directory, repetition_directories in experiment_directories.items():

        data = None

        trg_directory = os.path.join(experiment_directory, statistics_directory)
        exputils.io.makedirs(trg_directory)

        if verbose:
            print('Calculate statistics for {!r}:'.format(experiment_directory))

        for statistic_name, statistic_func in statistics:

            # calculate statistics if they do not exist
            filename_npy = '{}.npy'.format(statistic_name)
            filename_npz = '{}.npz'.format(statistic_name)

            filepath_npy = os.path.join(trg_directory, filename_npy)
            filepath_npz = os.path.join(trg_directory, filename_npz)
            if (not os.path.isfile(filepath_npy) and not os.path.isfile(filepath_npz)) or recalculate_statistics:

                if verbose:
                    print('\t{} ...'.format(statistic_name))

                data = get_data(data, repetition_directories)

                stat = statistic_func(statistic_name, data)

                if isinstance(stat, dict):
                    np.savez(filepath_npz, **stat)
                else:
                    np.save(filepath_npy, stat)


def load_data_from_repetitions(repetition_directories, data_subdirectory=None):
    '''Loads the data from all subdirectories of the given repetition directories.'''

    # use default logs directory to define where the data is
    if data_subdirectory is None: data_subdirectory = exputils.DEFAULT_DATA_DIRECTORY

    # collect data from each repetition
    data = dict()
    for repetition_directory in sorted(repetition_directories):

        # get id of the repetition from its foldername
        numbers_in_string = [int(s) for s in os.path.basename(repetition_directory).split('_') if s.isdigit()]
        repetition_id = numbers_in_string[0]

        data[repetition_id] = exputils.io.load_numpy_files(os.path.join(repetition_directory, data_subdirectory))

    return data



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import numpy as np
import plotly.subplots
from typing import Optional

# TODO: Bugfix - if sveral subfigures are shown, then the boxes are positioned in each plot as if they are plotted
#       in one subfigure

def plotly_box(data: Optional[list] = None,
               config: Optional[dict] = None,
               **kwargs):
    """
     Interactive box plot that shows several statistics such as the mean, std and range of scalars
     over all repetitions of each experiment.

    <figure markdown="span">
          ![plotly_box](../assets/images/plotly_box.png)
    </figure>

     Parameters:
         data (list): Data to plot.
         config (dict): Dictionary with configuration of plot.

     __Configuration__:

     - `layout` (`dict`): See [Plotly layout](https://plotly.com/python/reference/layout/) for all possible options.
         - `xaxis` (`dict`):
             - `title` (`str`): Title of the x-axis.
             - `range` (`tuple`): Tuple with min and max values of x-axis. Default is `[None, None]`.
         - `yaxis` (`dict`)
             - `title` (`str`): Title of the y-axis.
             - `range` (`tuple`): Tuple with min and max values of y-axis. Default is `[None, None]`.

     Returns:
         fig (figure): Plotly figure object that can be displayed using `display(fig)`.

     The plot is based on [Plotly box plots](https://plotly.com/python/box-plots/).
     """

    default_config = eu.AttrDict(
        subplots=eu.AttrDict(
            rows=None,
            cols=None,
            print_grid=False
        ),
        init_mode='mean_std',  # mean_std, mean, elements
        layout=eu.AttrDict(

            default_xaxis=eu.AttrDict(),  # if several subplots, then these are the default values for all xaxis config in fig.layout
            default_yaxis=eu.AttrDict(),  # if several subplots, then these are the default values for all yaxis config in fig.layout

            xaxis=eu.AttrDict(),
            yaxis=eu.AttrDict(),

            boxmode='group',

            updatemenus=[
                eu.AttrDict(
                    type="buttons",
                    active=0,
                    buttons=[
                         eu.AttrDict(label='normal',
                                     method='restyle',
                                     args=[{'boxpoints': False}]),
                         eu.AttrDict(label='all',
                              method='restyle',
                              args=[{'boxpoints': 'all'}]),
                         eu.AttrDict(label='suspectedoutliers',
                              method='restyle',
                              args=[{'boxpoints': 'suspectedoutliers'}]),
                         eu.AttrDict(label='outliers',
                              method='restyle',
                              args=[{'boxpoints': 'outliers'}]),
                     ],
                     direction='right',
                     pad={'t': 70},
                     x=1,
                     xanchor='right',
                     y=0,
                     yanchor='top'
                     ),
            ]
        ),

        default_trace=eu.AttrDict(
            legendgroup=None,
            boxmean='sd',
        ),
        default_subplot_traces=[],
        traces=[],

        labels=[],  # holds all labels in a specific structure

        default_trace_label='<trace_idx>',
        trace_labels=[],

        default_group_label='<group_idx>',
        group_labels=[],

        default_colors=plotly.colors.DEFAULT_PLOTLY_COLORS,
    )
    config = eu.combine_dicts(kwargs, config, default_config)

    if data is None:
        data = np.array([])

    # format data in form [subplot_idx:list][trace_idx:list][elems_per_trace:numpy.ndarray]
    if isinstance(data, np.ndarray):
        data = [[data]]
    elif isinstance(data, list) and isinstance(data[0], np.ndarray):
        data = [data]
    elif not isinstance(data, list) and not isinstance(data[0], list) and not isinstance(data[0][0], np.ndarray):
        raise ValueError('Unsupported type of data!')

    # handle different input formats of labels
    if config.labels:
        # if only labels for mean-traces are given, then add an empty label for the sub figure
        if isinstance(config.labels, list) and not isinstance(config.labels[0], tuple):
            config.labels = [('', config.labels)]
        # if no labels are given for elements, then create an empty list for element labels
        for ds_idx in range(len(config.labels)):
            for trace_idx in range(len(config.labels[ds_idx][1])):
                if not isinstance(config.labels[ds_idx][1][trace_idx], tuple):
                    config.labels[ds_idx][1][trace_idx] = (config.labels[ds_idx][1][trace_idx], [])

    # subplot_titles
    if config.labels and ('subplot_titles' not in config.subplots or config.subplots.subplot_titles == []):
        config.subplots.subplot_titles = [subplot_labels[0] for subplot_labels in config.labels]

    # identify the number of subplots
    n_subplots = len(data)

    # if not defined, set rows and cols of subplots
    if config.subplots.rows is None and config.subplots.cols is None:
        config.subplots.rows = n_subplots
        config.subplots.cols = 1
    elif config.subplots.rows is not None and config.subplots.cols is None:
        config.subplots.cols = int(np.ceil(n_subplots / config.subplots.rows))
    elif config.subplots.rows is None and config.subplots.cols is not None:
        config.subplots.rows = int(np.ceil(n_subplots / config.subplots.cols))

    # handle init mode
    if config.init_mode == 'normal':
        if len(config.layout.updatemenus) > 0:
            config.layout.updatemenus[0]['active'] = 0
        config.default_trace.boxpoints = False
    elif config.init_mode == 'all':
        if len(config.layout.updatemenus) > 0:
            config.layout.updatemenus[0]['active'] = 1
        config.default_trace.boxpoints = 'all'
    elif config.init_mode == 'suspectedoutliers':
        if len(config.layout.updatemenus) > 0:
            config.layout.updatemenus[0]['active'] = 2
        config.default_trace.boxpoints = 'suspectedoutliers'
    elif config.init_mode == 'outliers':
        if len(config.layout.updatemenus) > 0:
            config.layout.updatemenus[0]['active'] = 3
        config.default_trace.boxpoints = 'outliers'

    # make figure with subplots
    fig = plotly.subplots.make_subplots(**config.subplots)

    traces = []

    # interate over subplots
    for subplot_idx, subplot_data in enumerate(data):

        subplot_traces = []

        # create for each experiment a trace
        for trace_idx, cur_data in enumerate(subplot_data):

            data_points = np.array([])
            elem_labels = []

            if np.ndim(cur_data) == 0:
                data_points = np.array([cur_data])
                elem_labels = ['']
            elif np.ndim(cur_data) == 1:
                data_points = cur_data
                elem_labels = [''] * len(data_points)
            else:
                # collect data over elements
                for elem_idx, elem_data in enumerate(cur_data):  # data elements

                    # get element data which could be in matrix format or array format
                    if np.ndim(elem_data) == 0:
                        cur_elem_data = np.array([elem_data])
                    elif np.ndim(elem_data) == 1:
                        cur_elem_data = elem_data
                    elif np.ndim(elem_data) == 2:
                        if elem_data.shape[0] == 1:
                            cur_elem_data = elem_data[0, :]
                        elif elem_data.shape[1] == 1:
                            cur_elem_data = elem_data[1, 0]
                        else:
                            raise ValueError('Invalid data format!')
                    else:
                        raise ValueError('Invalid data format!')

                    data_points = np.hstack((data_points, cur_elem_data))

                    # handle trace for mean values
                    group_label = config.default_group_label
                    if len(config.group_labels) > elem_idx:
                        group_label = config.group_labels[elem_idx]
                    group_label = eu.misc.replace_str_from_dict(str(group_label), {'<group_idx>': elem_idx})

                    elem_labels.extend([group_label] * len(cur_elem_data))

            # handle trace for mean values
            if config.labels:
                trace_label = config.labels[subplot_idx][1][trace_idx][0]
            else:
                trace_label = config.default_trace_label
                if len(config.trace_labels) > trace_idx:
                    trace_label = config.trace_labels[trace_idx]
            trace_label = eu.misc.replace_str_from_dict(str(trace_label), {'<trace_idx>': trace_idx})

            trace_params = eu.AttrDict(
                x=elem_labels,
                y=data_points,
                name=trace_label,
                marker_color=config.default_colors[trace_idx % len(config.default_colors)])

            trace_config = config.default_trace.copy()
            if len(config.default_subplot_traces) > subplot_idx:
                trace_config = eu.combine_dicts(config.default_subplot_traces[subplot_idx], trace_config)
            if len(config.traces) > trace_idx:
                trace_config = eu.combine_dicts(config.traces[trace_idx], trace_config)

            trace_params = eu.combine_dicts(trace_config, trace_params)

            # handle legendgroup
            trace_legendgroup = trace_params.legendgroup
            if isinstance(trace_legendgroup, str):
                trace_legendgroup = eu.misc.replace_str_from_dict(trace_legendgroup,
                                                                  {'<trace_idx>': trace_idx,
                                                                   '<subplot_idx>': subplot_idx})
            trace_params.legendgroup = trace_legendgroup

            cur_trace = plotly.graph_objs.Box(**trace_params)
            subplot_traces.append(cur_trace)

        traces.append(subplot_traces)

    # set for the std toggle buttons which traces should be hidden and which ones should be shown
    layout = config.layout

    # set default values for all layouts
    def set_axis_properties_by_default(axis_name, fig_layout, config_layout):
        # sets the axis properties to default values

        def set_single_axis_property_default(cur_axis_name, default_name):
            if cur_axis_name in fig_layout or cur_axis_name in config_layout:
                cur_config = config_layout[cur_axis_name] if cur_axis_name in config_layout else dict()
                config_layout[cur_axis_name] = eu.combine_dicts(cur_config, config_layout[default_name])

        default_name = 'default_' + axis_name

        set_single_axis_property_default(axis_name, default_name)
        set_single_axis_property_default(axis_name + '1', default_name)
        axis_idx = 2
        while True:
            cur_axis_name = axis_name + str(axis_idx)

            if cur_axis_name not in fig_layout and cur_axis_name not in config_layout:
                break

            set_single_axis_property_default(cur_axis_name, default_name)
            axis_idx += 1

    set_axis_properties_by_default('xaxis', fig['layout'], layout)
    set_axis_properties_by_default('yaxis', fig['layout'], layout)

    # remove default fields, because they are not true proerties of the plotly layout
    del (layout['default_xaxis'])
    del (layout['default_yaxis'])

    cur_row = 1
    cur_col = 1
    for subplot_idx in range(n_subplots):
        n_traces = len(traces[subplot_idx])

        fig.add_traces(traces[subplot_idx],
                       rows=[cur_row] * n_traces,
                       cols=[cur_col] * n_traces)

        if cur_col < config.subplots.cols:
            cur_col += 1
        else:
            cur_col = 1
            cur_row += 1

    fig['layout'].update(layout)

    return fig


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
from exputils.misc.misc import numpy_vstack_2d_default
from exputils.misc.misc import replace_str_from_dict
from exputils.misc.misc import do_subdict_boolean_filtering
from exputils.misc.misc import get_dict_variable
from exputils.misc.misc import str_to_slices
from exputils.misc.misc import list_equal
from exputils.misc.misc import dict_equal
from exputils.misc.misc import moving_average
from exputils.misc.misc import call_function_from_config
from exputils.misc.misc import create_object_from_config
from exputils.misc.misc import seed
from exputils.misc.misc import is_allowed
from exputils.misc.misc import str_to_list
from exputils.misc.misc import get_experiment_name
from exputils.misc.misc import get_repetition_name
from exputils.misc.misc import mannwhitneyu_pvalue
from exputils.misc.misc import update_status



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
from exputils.misc.attrdict import AttrDict, combine_dicts
import exputils as eu
import numpy as np
import os
import copy
import warnings
from datetime import datetime
import re

# try to import tensorboard
try:
    import torch.utils.tensorboard
    is_exist_tensorboard_module = True
except ImportError:
    is_exist_tensorboard_module = False


def _get_safe_name(name):
    """Returns a name that is safe to save as a log entry."""
    return name.replace('/', '_')


class Logger:
    """
        Configuration:
            numpy_log_mode: String that defines how numpy data is logged.
                'npy': each property is loggend in an individual npy file
                'npz': all properties are combined in a npz file
                'cnpz': all properties are combined in a compressed npz file

            numpy_npz_filename: Name of the npz file if numpy data should be saved in a npz or compressed npz.
    """

    def default_config(self):
        dc = AttrDict(
            directory = None,
            numpy_log_mode = 'npy',
            numpy_npz_filename = 'logging.npz',

            tensorboard = AttrDict(
                log_dir = None,
                filename_suffix = '.tblog',
                purge_step=None,
                max_queue=10,
                flush_secs=120,
            )
        )
        return dc


    def __init__(self, config=None, **kwargs):
        self.config = combine_dicts(kwargs, config, self.default_config())

        self.numpy_data = dict()
        self.object_data = dict()

        self._tensorboard_writer = None
        self._is_tensorboard_active = False


    @property
    def directory(self):
        directory = self.config.directory
        if directory is None:
             directory = eu.DEFAULT_DATA_DIRECTORY
        return directory


    @directory.setter
    def directory(self, value):
        self.config.directory = value


    def __getitem__(self, key):
        key = _get_safe_name(key)

        if key in self.numpy_data:
            return self.numpy_data[key]
        elif key in self.object_data:
            return self.object_data[key]
        else:
            return None


    def __contains__(self, item):
        item = _get_safe_name(item)

        return (item in self.numpy_data) or (item in self.object_data)


    def items(self):
        return list(self.numpy_data.items()) + list(self.object_data.items())


    def clear(self, name=None):
        """
        Clears the data of the whole log or of a specific data element.

        :param name: If none, then the whole log is cleared, otherwise only the data element with the given name.
                     (default=None)
        """
        if name is not None:
            name = _get_safe_name(name)

        if name is None:
            self.numpy_data.clear()
            self.object_data.clear()
        else:
            if name not in self:
                raise ValueError('Unknown data element with name {!r}!'.format(name))

            if name in self.numpy_data:
                del self.numpy_data[name]

            if name in self.object_data:
                del self.numpy_data[name]


    def add_value(self, name, val, log_to_tb=None, tb_global_step=None, tb_walltime=None):

        safe_name = _get_safe_name(name)

        if safe_name not in self.numpy_data:
            self.numpy_data[safe_name] = []

        self.numpy_data[safe_name].append(val)

        if log_to_tb is True or (self._is_tensorboard_active and log_to_tb is not False):
            # identify if the value is a scalar, if yes, then add it to tensorboard
            if not isinstance(val, (list, tuple, np.ndarray)):
                self.tensorboard.add_scalar(name, val, tb_global_step, tb_walltime)
            else:
                warnings.warn('Can not log value for "{}" to tensorboard as it is not a scalar. Value: {}'.format(name, val))


    def add_scalar(self, name, scalar, log_to_tb=None, tb_global_step=None, tb_walltime=None):
        # same functionality as add_value, but is more consistent with the naming of tensorboard API
        self.add_value(name, scalar, log_to_tb=log_to_tb, tb_global_step=tb_global_step, tb_walltime=tb_walltime)


    def add_histogram(self, name, values, log_to_tb=None, tb_global_step=None, tb_walltime=None):

        safe_name = _get_safe_name(name)

        if safe_name not in self.numpy_data:
            self.numpy_data[safe_name] = []

        self.numpy_data[safe_name].append(values)

        if log_to_tb is True or (self._is_tensorboard_active and log_to_tb is not False):
            # values must be a numpy array
            values = np.array(values)
            self.tensorboard.add_histogram(name, values, global_step=tb_global_step, walltime=tb_walltime)


    def add_object(self, name, obj):
        """
        Adds an object to the log that will be saved in a dill file when the log is saved.

        :param name:
        :param obj:
        :return:
        """
        name = _get_safe_name(name)

        if name not in self.object_data:
            self.object_data[name] = []

        self.object_data[name].append(copy.deepcopy(obj))


    def add_single_object(self, name, obj, directory=None):
        """
        Adds a single object to the log by directly writing it to a file.
        Overwrites existing object data with the same name.
        """
        if directory is None:
            directory = self.directory

        name = _get_safe_name(name)

        file_path = os.path.join(directory, name)
        eu.io.save_dill(obj, file_path)


    def load_single_object(self, name):
        """
        Loads a single object from a file.

        :param name: Name of the object.
        :return: Loaded object.
        """
        name = _get_safe_name(name)

        file_path = os.path.join(self.directory, name)
        return eu.io.load_dill(file_path)


    def save(self, directory=None):
        directory = self.directory if directory is None else directory

        if directory is None:
            raise ValueError('A directory in which the log will be saved must be provided!')

        # make sure the dsirectory exists
        eu.io.makedirs(directory)

        # numpy data

        if self.config.numpy_log_mode.lower() == 'npy':
            path = directory
        else:
            path = os.path.join(directory, self.config.numpy_npz_filename)

        eu.io.save_dict_to_numpy_files(self.numpy_data, path, self.config.numpy_log_mode)

        # object data
        for obj_name, obj in self.object_data.items():
            file_path = os.path.join(directory, obj_name)
            eu.io.save_dill(obj, file_path)

        # save also tensorboard if one exists
        if self._tensorboard_writer is not None:
            self._tensorboard_writer.flush()


    def load(self, directory=None, load_objects=False):
        directory = self.directory if directory is None else directory

        if directory is None:
            raise ValueError('A directory in which the log will be saved must be provided!')

        self.numpy_data = eu.io.load_numpy_files(directory)

        # in the case that all data was logged into a npz file
        if len(self.numpy_data) == 1 and 'logging' in self.numpy_data:
            self.numpy_data = self.numpy_data['logging']

        for key, item in self.numpy_data.items():
            self.numpy_data[key] = item.tolist()

        if load_objects:
            self.object_data = eu.io.load_dill_files(directory)
        else:
            self.object_data = dict()


    @property
    def is_tensorboard_active(self):
        """Return True if a tensorboard is active and can be used, otherwise False."""
        return self._is_tensorboard_active


    @property
    def tensorboard(self):
        """Tensorboard SummaryWriter."""

        if self._tensorboard_writer is None:
            self.create_tensorboard()

        return self._tensorboard_writer


    def create_tensorboard(self, config=None, **kwargs):
        """Creates a tensorboard"""

        if not is_exist_tensorboard_module:
            raise ImportError('Tensorboard module torch.utils.tensorboard does not exist!')

        self.config.tensorboard = eu.combine_dicts(kwargs, config, self.config.tensorboard)

        # identify which experiment and repetition we are in, to set the tensorboard folder
        experiment_name = eu.misc.get_experiment_name()
        repetition_name = eu.misc.get_repetition_name()

        if self.config.tensorboard.log_dir is None:
            # create the logs in the experiment folder which is located on top of the repetition and experiment folder if they exist

            log_dir = ''

            if experiment_name is not None:
                log_dir = os.path.join(log_dir, '..')

            if repetition_name is not None:
                log_dir = os.path.join(log_dir, '..')

            log_dir = os.path.join(log_dir, 'tensorboard_logs')

            self.config.tensorboard.log_dir = log_dir

        # add the experiment and repetition name to the log path, to create sub logs for them
        if experiment_name is not None:
            experiment_name = re.sub('_0{1,5}', '_', experiment_name)
            experiment_name = experiment_name.replace('experiment', 'exp')
            self.config.tensorboard.log_dir = os.path.join(self.config.tensorboard.log_dir, experiment_name)
        if repetition_name is not None:
            repetition_name = re.sub('_0{1,5}', '_', repetition_name)
            repetition_name = repetition_name.replace('experiment', 'exp')
            self.config.tensorboard.log_dir = os.path.join(self.config.tensorboard.log_dir, repetition_name.replace('repetition', 'rep'))

        dt = datetime.now()
        self.config.tensorboard.log_dir = os.path.join(self.config.tensorboard.log_dir, dt.strftime('%y.%m.%d_%H.%M'))

        if self._tensorboard_writer is not None:
            self._tensorboard_writer.flush()
            warnings.warn('Tensorboard SummaryWriter existed already. Creating a new one ...')

        self._tensorboard_writer = torch.utils.tensorboard.SummaryWriter(**self.config.tensorboard)

        return self._tensorboard_writer


    def activate_tensorboard(self, config=None, **kwargs):
        """Activates the tensorboard to automatically also log values that are given to the log.
        If not tensorboard exists, one is created."""

        if self._tensorboard_writer is None:
            self.create_tensorboard(config=config, **kwargs)

        self._is_tensorboard_active = True

        return self._tensorboard_writer


    def deactivate_tensorboard(self):
        """Deactivates the tensorboard to automatically also log values that are given to the log."""
        if self._is_tensorboard_active:
            self._is_tensorboard_active = False
            self._tensorboard_writer.flush()



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
from .loading import load_experiment_descriptions
from .loading import load_experiment_data
from .loading import load_single_experiment_data
from .loading import load_experiment_python_module
from .loading import load_experiment_data_single_object
from .selection import select_experiment_data
from .statistics import calc_repetition_statistics
from .statistics import calc_statistics_over_repetitions
from .utils import get_ordered_experiment_ids_from_descriptions
from .logger import Logger
import exputils.data.logging










##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import warnings
import numpy as np

# TODO: Feature - output_formats
# TODO: Feature - data_filter
# TODO: Feature - data_filter_inds

def select_experiment_data(experiment_data, datasources, experiment_ids='all', repetition_ids='all', output_format=('S', 'E', 'D'), data_filter=None, data_filter_inds=None, experiment_descriptions=None, config=None, **kwargs):
    '''
    Collects the data for specific datasources, experiments and repetitions from the experiment data dictionary.
    The output format of the collected data can be chosen.
    Also allows to filter data.


    Output format: 'S' - datasource, 'E' - experiment, 'D' - repetition/data

    Default: ('S','E','D')

    :param datasources: String or list of strings with specification of the data sources from which the data should be collected.
    :param experiment_ids: Experiment id, List of experiment ids, or 'all'. (Default: 'all')
    :param repetition_ids: Repetition id, List of repetition ids, or 'all', or 'none'. (Default: 'all')
                           'none' means that the data is not over repetitions.
    '''

    S, E, D = 'S', 'E', 'D'

    default_config = eu.AttrDict(
        datasource_label='<datasource>', # either string with template for all data sources or a list with a string for each label
        experiment_label='<name>',
        repetition_label='<short_name> - <repetition_id>',
    )
    config = eu.combine_dicts(kwargs, config, default_config)

    if not experiment_data:
        return []

    # handle input parameters

    # get data sources as a list
    if not isinstance(datasources, list):
        datasources = [datasources]

    if not isinstance(config.datasource_label, list):
        config.datasource_label = [config.datasource_label] * len(datasources)

    # get experiment_ids as a list of ids
    if experiment_ids is None:
        experiment_ids = ['all']
    elif not isinstance(experiment_ids, list):
        experiment_ids = [experiment_ids]
    if experiment_ids == ['all']:
        experiment_ids = list(experiment_data.keys())

    # get repetition_ids as a list of ids or a slice over all repetitions
    if repetition_ids is None:
        repetition_ids = ['none']
    elif not isinstance(repetition_ids, list):
        repetition_ids = [repetition_ids]


    # if data_filter is not None and data_filter:
    #     # go over all experiments and apply the filter to each of them individually
    #     experiment_filter_inds = dict()
    #
    #     for exp_id in experiment_ids:
    #
    #     # filter data according data_filter the given filter
    #         data_filter_inds = filter_single_experiment_data(rep_data, data_filter)
    #
    #         experiment_filter_inds[exp_id] = ...


    # collect the data for each datasource and experiment
    collected_data = []
    data_labels = []
    for datasource_idx, datasource in enumerate(datasources):

        is_transpose = False
        if datasource.endswith('\''):
            datasource = datasource[:-1]
            is_transpose = True

        # identify the type of data
        # 3 types exist:
        #   - exp_data: data directly for an experiment, does not contain single dataitems for repetitions
        #   - rep_data: data for each repetition
        #   - exp_rep_data: data under the experiment, that whos first data dimension is over repetitions
        if datasource.startswith('exp.'):
            if repetition_ids == 'none':
                data_type = 'exp_data'
            else:
                data_type = 'exp_rep_data'
        elif datasource.startswith('rep.'):
            data_type = 'rep_data'
        else:
            # check datatype automatically
            # first check if the data is in experiment data
            try:
                # try if the data can be loaded from the experiment data (check for first experiment)
                eu.misc.get_dict_variable(experiment_data[experiment_ids[0]], datasource)
                # yes --> experiment data
                if repetition_ids == ['none']:
                    data_type = 'exp_data'
                else:
                    data_type = 'exp_rep_data'

            except (KeyError, IndexError):
                # could not find datasource under experiment_data, thus assume it is under repetition_data
                data_type = 'rep_data'

        if data_type in ['rep_data', 'exp_rep_data'] and repetition_ids == ['none']:
            raise ValueError('Data for datasource {!r} is assumed to be of data_type {!r}. In this case it is not possible that repetition_ids are \'none\'!'.format(datasource, data_type))

        cur_experiments_data = []
        cur_experiments_labels = []

        for experiment_id in experiment_ids:

            rep_ids = []

            # check if the data exisits
            if experiment_id not in experiment_data:
                warnings.warn('Data {!r} for experiment {!r} does not exist! Data is set to None.'.format(datasource,
                                                                                                          experiment_id))
                cur_data = None
            else:

                if data_type == 'exp_data':
                    rep_ids = []
                    try:
                        cur_data = eu.misc.get_dict_variable(experiment_data[experiment_id], datasource)
                    except (KeyError, IndexError):
                        # data does not exists
                        warnings.warn('Data {!r} for experiment {!r} does not exist! Data is set to None.'.format(datasource,
                                                                                                                  experiment_id))
                        cur_data = None


                elif data_type == 'exp_rep_data':

                    # if all repetition ids should be taken, then use a slice operator
                    rep_ids = repetition_ids.copy()
                    if rep_ids == ['all']:
                        rep_ids = slice(None)

                    try:
                        cur_data = eu.misc.get_dict_variable(experiment_data[experiment_id], datasource)
                        cur_data = cur_data[rep_ids]
                    except (KeyError, IndexError):
                        # data does not exists
                        warnings.warn('Data {!r} for experiment {!r} does not exist! Data is set to None.'.format(datasource,
                                                                                                                  experiment_id))
                        cur_data = None

                elif data_type == 'rep_data':

                    # how to store the data depends on the datatype:
                    #   scalar --> 1D numpy array
                    #   nD numpy array --> (n+1)D numpy array with first dimension over repetitions
                    #   other object --> list

                    cur_repetition_data = experiment_data[experiment_id]['repetition_data']

                    rep_ids = repetition_ids.copy()
                    if rep_ids == ['all']:
                        n_loaded_repetitions = len(cur_repetition_data)
                        rep_ids = list(range(n_loaded_repetitions))

                    # go over each repetition and store data in a list
                    cur_data_per_rep = []
                    is_numpy_array_type = True
                    final_np_array_shape = []
                    is_data_exist_inds = []
                    for rep_id in rep_ids:

                        try:
                            cur_rep_data = eu.misc.get_dict_variable(cur_repetition_data[rep_id], datasource)
                            is_data_exist_inds.append(True)
                        except (KeyError, IndexError):
                            # data does not exists in this repetition
                            warnings.warn('Data {!r} for repetition {!r} of experiment {!r} does not exist.'.format(datasource, rep_id, experiment_id))
                            cur_rep_data = None
                            is_data_exist_inds.append(False)

                        cur_data_per_rep.append(cur_rep_data)

                        # detetct if the data can be put into an numpy array,
                        # and detect the maximum shape of the array
                        if is_numpy_array_type and is_data_exist_inds[-1]:

                            cur_np_array_shape = None
                            if np.isscalar(cur_rep_data):
                                cur_np_array_shape = [1]
                            elif isinstance(cur_rep_data, np.ndarray):
                                cur_np_array_shape = cur_rep_data.shape
                            else:
                                is_numpy_array_type = False

                            if cur_np_array_shape is not None:
                                # check if all data have same number of dimensions
                                if not final_np_array_shape:
                                    final_np_array_shape = cur_np_array_shape
                                elif len(final_np_array_shape) == len(cur_np_array_shape):
                                    final_np_array_shape = np.maximum(final_np_array_shape, cur_np_array_shape)
                                else:
                                    is_numpy_array_type = False

                    if is_numpy_array_type:
                        # if the data can be transformed into a numpy array
                        if len(final_np_array_shape) == 1 and final_np_array_shape[0] == 1:
                            # if the data per repetition is only a scalar, then do not create
                            # create an extra dimensions for it, this replicates the default beahvior of numpy
                            data_shape = [len(cur_data_per_rep)]
                        else:
                            data_shape = [len(cur_data_per_rep)] + list(final_np_array_shape)

                        cur_data = np.full(data_shape, np.nan)
                        for rep_idx, rep_data in enumerate(cur_data_per_rep):

                            # only set values if data exisited for the repetition
                            if is_data_exist_inds[rep_idx]:

                                # create the correct slicing to add the rep_data into the whole array
                                rep_data_shape = np.shape(rep_data)
                                if len(data_shape) == 1:
                                    slices = rep_idx
                                else:
                                    slices = tuple([rep_idx] + [slice(0, d) for d in rep_data_shape])

                                cur_data[slices] = rep_data
                    else:
                        # otherwise keep data in list form over repetitions
                        cur_data = cur_data_per_rep

                    # set data to None if no repetition had some data
                    if not np.any(is_data_exist_inds):
                        cur_data = None

            if is_transpose:
                cur_data = np.transpose(cur_data)

            cur_experiments_data.append(cur_data)

            # create label string for the experiment
            exp_str_replace_dict = {'experiment_id': experiment_id,
                                    'datasource': datasource}
            if experiment_descriptions is not None:
                exp_str_replace_dict['name'] = experiment_descriptions[experiment_id]['name']
                exp_str_replace_dict['short_name'] = experiment_descriptions[experiment_id]['short_name']
                exp_str_replace_dict['description'] = experiment_descriptions[experiment_id]['description']
                exp_str_replace_dict['directory'] = experiment_descriptions[experiment_id]['directory']

            experiment_label = eu.misc.replace_str_from_dict(
                config.experiment_label,
                exp_str_replace_dict,
                pattern_format='<{}>')

            # create label strings for each repetition
            repetition_labels = []
            if rep_ids:
                rep_str_replace_dict = {'experiment_id': experiment_id,
                                        'datasource': datasource}
                if experiment_descriptions is not None:
                    rep_str_replace_dict['name'] = experiment_descriptions[experiment_id]['name']
                    rep_str_replace_dict['short_name'] = experiment_descriptions[experiment_id]['short_name']
                    rep_str_replace_dict['description'] = experiment_descriptions[experiment_id]['description']
                    rep_str_replace_dict['directory'] = experiment_descriptions[experiment_id]['directory']

                if isinstance(rep_ids, slice):
                    # assume it is a slice(None), i.e. a slice over all possible repetitions
                    # then create ids
                    rep_ids = list(range(len(cur_data)))

                for rep_id in rep_ids:
                    rep_str_replace_dict['repetition_id'] = rep_id
                    repetition_label = eu.misc.replace_str_from_dict(
                        config.repetition_label,
                        rep_str_replace_dict,
                        pattern_format='<{}>')
                    repetition_labels.append(repetition_label)

            if repetition_labels:
                cur_experiments_labels.append((experiment_label, repetition_labels))
            else:
                cur_experiments_labels.append(experiment_label)

        collected_data.append(cur_experiments_data)

        # define the data source label
        datasource_label = eu.misc.replace_str_from_dict(
            config.datasource_label[datasource_idx],
            {'<datasource>': datasource})
        data_labels.append((datasource_label, cur_experiments_labels))


    # reformat the data according to the given data format
    if output_format != (S, E, D):
        raise NotImplementedError('Output format {!r} is not supported!'.format(output_format))
        #
        # reformated_data = []
        #
        # cur_level_data = []
        # for cur_data_level_format in output_format:
        #
        #     # just encapsulating list of 1 is given
        #     if cur_data_level_format == 1:
        #         cur_level_data.append([])
        #     else:
        #         pass


    return collected_data, data_labels

#
# def recursive_reformat_data(collected_data, output_format, SED_indexes=None):
#
#     if SED_indexes is None:
#         SED_indexes = [None, None, None] # [data source, experiment, repetition]
#
#     # if last level, then create a numpy array
#
#     def get_data_object_combinations(cur_level_format, SED_indexes):
#         S, E, D = 'S', 'E', 'D'
#         S_loc, E_loc, D_loc = 0, 1, 2
#
#         # does the current data level has a parent or not
#         # if yes, then take all possibilities
#         # if no, then just take the exisiting possibilities under the parent
#
#         # return list of tuples with DER
#
#         # if parent is a datasource, then
#
#         # identify the current combinations:
#         data_type_combinations = get_cur_level_format_combinations(cur_level_format) # 'ExD' --> ['E', 'D']
#
#         for data_type in data_type_combinations:
#
#             if data_type == S:
#
#                 if SED_indexes[S_loc] is not None:
#                     raise ValueError('Each data type can only used once! \'S\' was used several times.')
#
#                 if SED_indexes[E_loc] is None and SED_indexes[D_loc] is None:
#                     # get all data sources
#                      data_object_combinations = np.arange(len(collected_data))
#
#                 else:
#
#                     data_object_combinations = []
#
#                     if SED_indexes[E_loc] is not None:
#                         # look if the given experiment has data for the datasource
#                         for ds_idx in range(len(collected_data)):
#                             if __name__ == '__main__':
#                                 if collected_data[ds_idx][SED_indexes[E_loc]] ..
#
#
#                 pass
#             elif data_type == E:
#                 pass
#             elif data_type == D:
#                 pass
#             else:
#                 raise ValueError('Unknown data type')
#
#
#         return data_object_combinations
#
#
#
#
#
#     cur_level_format = output_format[0]
#
#     # just encapsulating list if 1 is given
#     if cur_level_format == 1:
#
#         if len(output_format) <= 1:
#             raise ValueError('Final element in output format can not be \'1\'!')
#
#         cur_data = recursive_reformat_data(collected_data,
#                                            output_format[1:],
#                                            SED_indexes)
#         reformated_data = [cur_data]
#
#     else:
#         # otherwise, identify which data should be represented by this level
#
#         data_object_combinations = get_data_object_combinations(cur_level_format, SED_indexes)
#
#         if len(output_format) > 1:
#             # check if last element is a slice over repetitions,
#             # if yes, then just grap them easily
#             reformated_data = np.array([])
#         else:
#             reformated_data = []
#
#         # go through the list data objects that are specified for this level
#         # could be data sources, experiments, repetitions or combinations of these
#         for data_object_combination in data_object_combinations:
#
#             # fill out the DER_indexes according to the given combinations:
#             if data_object_combination[0] is not None: SED_indexes[0] = data_object_combination[0]
#             if data_object_combination[1] is not None: SED_indexes[1] = data_object_combination[1]
#             if data_object_combination[2] is not None: SED_indexes[2] = data_object_combination[2]
#
#             if len(output_format) <= 1:
#                 # final level, get the data
#
#                 if len(data_object_combination) > 1:
#                     raise AssertionError('For the final level the number of data_object_combinations must be 1!')
#
#                 reformated_data = collected_data[SED_indexes[0]][SED_indexes[1]][SED_indexes[2]]
#
#             else:
#                 # not the final level, go one level down
#                 cur_data = recursive_reformat_data(collected_data,
#                                                    output_format[1:],
#                                                    SED_indexes)
#                 reformated_data.append(cur_data)
#
#
#     return reformated_data



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import warnings
from typing import Optional, Union

import numpy as np
import exputils as eu
import re
import os
import copy
import random
import scipy.stats
from datetime import datetime
from exputils.misc.attrdict import combine_dicts
# try to import torch, so that its seed can be set by the seed() - function
try:
    import torch
except ImportError:
    torch = None


def numpy_vstack_2d_default(array1, array2, default_value=np.nan):

    if len(array1) == 0:
        return array2

    if len(array2) == 0:
        return array1

    if np.ndim(array1) == 1:
        array1 = np.reshape(array1, (1,len(array1)))

    if np.ndim(array2) == 1:
        array2 = np.reshape(array2, (1,len(array2)))

    shape1 = np.shape(array1)
    shape2 = np.shape(array2)

    if shape1[1] == shape2[1]:
        return np.vstack((array1, array2))

    elif shape1[1] > shape2[1]:
        # add default values to array1

        new_values = np.ones((shape2[0], shape1[1] - shape2[1]))
        new_values[:] = default_value

        return np.vstack((array1, np.hstack((array2, new_values))))

    else:
        # add default values to array2

        new_values = np.ones((shape1[0], shape2[1] - shape1[1]))
        new_values[:] = default_value

        return np.vstack((np.hstack((array1, new_values)), array2))



def dict_equal(dict1, dict2):
    """Checks if two dictionaries are equal. Allows to check numpy arrays as content."""

    if not isinstance(dict1, dict) and not isinstance(dict2, dict):
        return dict1 == dict2

    if isinstance(dict1, dict) and not isinstance(dict2, dict):
        return False

    if not isinstance(dict1, dict) and isinstance(dict2, dict):
        return False

    # compare if set of keys is the same
    if set(dict1.keys()) != set(dict2.keys()):
        return False

    # compare all values
    for key in dict1.keys():

        # use special comparison for numpy arrays
        if isinstance(dict1[key], np.ndarray):
            if not np.array_equal(dict1[key], dict2[key]):
                return False
        else:
            if dict1[key] != dict2[key]:
                return False

    return True


def list_equal(list_a, list_b):
    """
    Checks if the content of two lists are equal. Can also handle numpy arrays as content.

    :param list_a: List 1.
    :param list_b: List 2.
    :return: True if the content of both lists are equal, otherwise False.
    """

    if len(list_a) != len(list_b):
        return False

    for idx in range(len(list_a)):

        if type(list_a[idx]) != type(list_b[idx]):
            return False
        if isinstance(list_a[idx], list):
            if not list_equal(list_a[idx], list_b[idx]):
                return False
        elif isinstance(list_a[idx], tuple):
            if not list_equal(list_a[idx], list_b[idx]):
                return False
        elif isinstance(list_a[idx], np.ndarray):
            if not np.array_equal(list_a[idx], list_b[idx], equal_nan=True):
                return False
        elif list_a[idx] != list_b[idx]:
            return False

    return True



def replace_str_from_dict(string, dictionary, pattern_format='{}'):
    """Replaces string occurances that are given as a dictionary."""
    out_string = string

    for key_name, new_str in dictionary.items():
        out_string = out_string.replace(pattern_format.format(key_name),
                                        '{}'.format(new_str))

    return out_string



def str_to_slices(slices_str):
    """
    Creates a slices for lists and numpy arrays from a string.

    >>> out = str_to_slices('[:,-1,0]')
    >>> print(out) # [slice(None), slice(-1), slice(0)]

    :param slice_str: String that describes the slices.
    :return: List of slices.
    """

    # remove all spaces
    slices_str = slices_str.replace(' ', '')

    if slices_str[0] != '[' or slices_str[-1] != ']':
        raise ValueError('Wrong slice format given. Needs to start witt\'[\' and end with \']\'!')

    # seperate each sub slice
    slice_strings = []
    cur_slice_str = ''
    is_inner_idx_list = False
    for c in slices_str[1:-1]:
        if c == ',' and not is_inner_idx_list:
            slice_strings.append(cur_slice_str)
            cur_slice_str = ''
        elif c == '[':
            is_inner_idx_list = True
            cur_slice_str += c
        elif c == ']':
            is_inner_idx_list = False
            cur_slice_str += c
        else:
            cur_slice_str += c
    slice_strings.append(cur_slice_str)

    slices = []
    for slice_str in slice_strings:

        if slice_str == ':':
            # slice_str: '[:]'
            slices.append(slice(None))
        elif ':' in slice_str:
            # slice_str: '[3:]' or '[3:5]' or '[3:10:2]'
            slice_params = []
            slice_str_params = slice_str.split(':')
            for str_p in slice_str_params:
                if str_p != '':
                    slice_params.append(int(str_p))
                else:
                    slice_params.append(None)

            slices.append(slice(*slice_params))
        else:
            if slice_str[0] == '[' and slice_str[-1] == ']':
                # list with indexes for numpy array indexing, e.g. '[[1,2,4]]'
                idxs = []
                for idx_str in slice_str[1:-1].split(','):
                   if idx_str != '':
                       idxs.append(int(idx_str))
                slices.append(idxs)
            else:
                # slice_str: '[3]'
                slices.append(int(slice_str))

    return slices


def get_dict_variable(base_dict, variable_str):
    """
    Retrieves items from sub-dictionaries in a dictionary using a single string.

    >>> d = {'sub_dict_1': {'sub_dict_2': {'item_1': 0}}}
    >>> print(get_dict_variable(d, 'sub_dict_1.sub_dict_2.item_1'))

    Allows also to index items in lists.

    >>> d = {'sub_dict_1': {'item_list': ['a', 'b', 'c']}}
    >>> print(get_dict_variable(d, 'sub_dict_1.item_list[0]'))
    >>> print(get_dict_variable(d, 'sub_dict_1.item_list[-1]'))

    :param base_dict: Dictionary with sub-dictionaries.
    :param variable_str: Path to item. Uses '.' to split sub dictionary keys and the item key.
    :return: Item value.
    """

    # TODO: Feature - allow lists of lists, e.g. 'sub_var.var[:][1]'

    # get the string describing the first sub element in the given variable string
    variable_subelement_strings = variable_str.split('.')
    cur_subelement_str = variable_subelement_strings[0]

    # check if the sub element string contains slices:
    if '[' in cur_subelement_str:
        slice_start = cur_subelement_str.find('[')

        sub_var_name = cur_subelement_str[:slice_start]
        slices = str_to_slices(cur_subelement_str[slice_start:])

        cur_value = base_dict[sub_var_name]

        if isinstance(cur_value, np.ndarray):
            cur_value = cur_value[tuple(slices)]
        else:
            for slice_obj in slices:
                try:
                    cur_value = cur_value[slice_obj]
                except TypeError:
                    raise IndexError()
    else:
        cur_value = base_dict[cur_subelement_str]

    if len(variable_subelement_strings) > 1:
        # not all sub elements processed yet
        cur_value = get_dict_variable(cur_value, '.'.join(variable_subelement_strings[1:]))

    return cur_value



def do_subdict_boolean_filtering(data, filter):
    """
    Filters a list of dictionaries using conditions based on items in the dictionaries.

    >>> d = [dict(x=1, y=1), dict(x=2, y=3), dict(x=3, y=9))]
    >>> filter1 = do_subdict_boolean_filtering(d, ('x', '==', 1))
    >>> filter2 = do_subdict_boolean_filtering(d, ('x', '==', 'y'))

    :param data: List with dictionaries.
    :param filter: Tuple with filter condition.
    :return: Boolean indices.
    """

    if isinstance(filter, tuple):

        if len(filter) == 3:

            bool_component_1 = do_subdict_boolean_filtering(data, filter[0])
            bool_component_2 = do_subdict_boolean_filtering(data, filter[2])

            if filter[1] == 'and':
                ret_val = bool_component_1 & bool_component_2
            elif filter[1] == 'or':
                ret_val = bool_component_1 | bool_component_2
            elif filter[1] == '<':
                ret_val = bool_component_1 < bool_component_2
            elif filter[1] == '<=':
                ret_val = bool_component_1 <= bool_component_2
            elif filter[1] == '>':
                ret_val = bool_component_1 > bool_component_2
            elif filter[1] == '>=':
                ret_val = bool_component_1 >= bool_component_2
            elif filter[1] == '==':
                ret_val = bool_component_1 == bool_component_2
            elif filter[1] == '!=':
                ret_val = bool_component_1 != bool_component_2
            elif filter[1] == '+':
                ret_val = bool_component_1 + bool_component_2
            elif filter[1] == '-':
                ret_val = bool_component_1 - bool_component_2
            elif filter[1] == '*':
                ret_val = bool_component_1 * bool_component_2
            elif filter[1] == '/':
                ret_val = bool_component_1 / bool_component_2
            elif filter[1] == '%':
                ret_val = bool_component_1 % bool_component_2
            else:
                raise ValueError('Unknown operator {!r}!'.format(filter[1]))

        elif len(filter) == 2:

            val_component_1 = do_subdict_boolean_filtering(data, filter[1])

            if filter[0] == 'sum':
                ret_val = np.sum(val_component_1)
            elif filter[0] == 'cumsum':
                ret_val = np.cumsum(val_component_1)
            elif filter[0] == 'max':
                ret_val = np.max(val_component_1)
            elif filter[0] == 'min':
                ret_val = np.min(val_component_1)
            else:
                raise ValueError('Unknown operator {!r}!'.format(filter[0]))

        else:
            raise ValueError('Unknown filter command {!r}!'.format(filter))

    else:

        is_var = False
        if isinstance(filter, str):

            # check if string is a variable in the data
            is_var = True
            try:
                # try if the string can be used to find a variable in the data
                # if yes --> the string is a variable
                # if no, these expressions will result in an error and the string is not a variable but some value
                if isinstance(data, list) or isinstance(data, np.ndarray):
                    get_dict_variable(data[0], filter)
                else:
                    # check first item if the data object has a __iter__ method such as the explorationdatahandler
                    for item in data:
                        get_dict_variable(item, filter)
                        break

            except KeyError:
                is_var = False

        if is_var:
            # if the string is a variable then get the data of the variable:
            ret_val = np.zeros(len(data))

            for data_idx, cur_data in enumerate(data):
                ret_val[data_idx] = get_dict_variable(cur_data, filter)
        else:
            ret_val = filter

    return ret_val


def moving_average(data, n, mode='fill_start'):

    data = np.array(data)

    if data.ndim == 1:

        ret = np.cumsum(data, dtype=float)
        ret[n:] = ret[n:] - ret[:-n]
        moving_mean = ret[n - 1:] / n

        if mode == 'fill_start':
            moving_mean =  np.hstack((np.full(n-1, moving_mean[0]), moving_mean))

    elif data.ndim == 2:

        ret = np.cumsum(data, dtype=float, axis=1)
        ret[:, n:] = ret[:, n:] - ret[:, :-n]
        moving_mean = ret[:, n - 1:] / n

        if mode == 'fill_start':
            moving_mean = np.hstack((np.tile(np.transpose([moving_mean[:,0]]), (1, n-1)), moving_mean))

    else:
        raise ValueError('Can compute the moving average only for arrays of dimension 1 or 2!')

    return moving_mean


def call_function_from_config(config, *args, func_attribute_name='func', **argv):
    """Calls a function that is defined as a config dictionary or AttrDict.

    The configuration dictionary must contain a property that holds the function handle. This is
    by default called `func`, but can be differently defined using the `func_attribute_name` parameter.
    All other properties of the configuration dictionary are used as parameters for the function call.

    If the given `config` argument is a function handle, then this function is called and the given
    *args and **argvs given as arguments.
    If the given `config` argument is not a dictionary or a function handle, then it is directly
    returned.

    Example:
        ```python
        import exputils as eu

        def calc_area(length, width, unit='sm'):
            area = length * width
            return f"{area} {unit}"

        config = eu.AttrDict()
        config.func = calc_area  # function that should be called
        config.unit = 'square meters'

        # calls the function with: calc_area(length=3, width=4, unit='square meters')
        out = eu.call_function_from_config(config, length=3, width=4)

        print(out)
        ```
        Output:
        ```
        '12 square meters'
        ```

    Parameters:
        config (dict): Configuration dictionary func property that holds the function handle.
        func_attribute_name (str): Name of the func attribute.
        *args: Additional arguments to pass to the function.
        *argv: Additional arguments to pass to the function.

    Returns:
        res (Any): The return value of the function.
    """

    if isinstance(config, dict) and func_attribute_name in config:

        func_handle = config[func_attribute_name]

        function_arguments = copy.deepcopy(config)
        del function_arguments[func_attribute_name]
        function_arguments = combine_dicts(argv, function_arguments)

        return func_handle(*args, **function_arguments)

    elif callable(config):
        return config(*args, **argv)

    else:
        return config


def create_object_from_config(config: dict, *args, **argv):
    """
    Creates a class object that is defined as a config dictionary or AttrDict.

    The configuration dictionary must contain a 'cls' property that holds the class type.
    All other properties are used as parameters for the constructor of the object.

    Example:
        ```python
        import exputils as eu
        from collections import Counter

        config = eu.AttrDict()
        config.cls = Counter  # class type that should be created
        config.green=2
        config.blue=1

        # creates the Counter object using: obj = Counter(red=3, green=2, blue=1)
        obj = eu.create_object_from_config(config, red=3)

        print(obj)
        ```
        Output:
        ```
        Counter({'red': 3, 'green': 2, 'blue': 1})
        ```

    Parameters:
        config (dict): Configuration dictionary with `cls` property that holds the class type.
        *args: Additional arguments to pass to the constructor of the object.
        *argv: Additional arguments to pass to the constructor of the object.

    Returns:
        obj (object): The resulting object.
    """
    return call_function_from_config(config, *args, func_attribute_name='cls', **argv)


def seed(seed: Optional[Union[int, dict]] = None,
         is_set_random: bool = True,
         is_set_numpy: bool = True,
         is_set_torch: bool = True) -> int:
    """
    Sets the random seed for random, numpy and pytorch (if it is installed).

    Parameters:
        seed (int, dict):
            Seed (integer) or a configuration dictionary which contains a 'seed' property.
            If `None` is given, a random seed is chosen.
        is_set_random (bool):
            Should the random seed of the python `random` package be set.
        is_set_numpy (bool):
            Should random seed of `numpy.random` be set.
        is_set_torch:
            Should random seed of torch be set.

    Returns:
        seed (int): Integer that was used as seed.
    """

    if seed is None:
        if torch:
            seed = torch.seed()
        else:
            seed = np.randint(2**32)

    elif isinstance(seed, dict):
        seed = seed.get('seed', None)

    if is_set_numpy:
        np.random.seed(seed)

    if is_set_random:
        random.seed(seed)

    if torch:
        if is_set_torch:
            torch.manual_seed(seed)

    return seed


def is_allowed(name, allowed_list=None, denied_list=None):
    """
    Checks if an entity (string name) is allowed based on either a list of allowed entities or denied entities.
    Only accepts either a list if allowed entities, or a list if denied entities, but not both.
    If neither list is provided then all entities are allowed.
    If the allowed entities list is provided, then only entities on this list are allowed.
    If the denied entities list is provided, then only entities that are not on this list are allowed.

    :param name: Name of entity (string)
    :param allowed_list: List of strings with allowed entities. If None then it is ignored. (default=None)
    :param denied_list: List of strings with denied entities. If None then it is ignored. (default=None)
    :return: True if allowed, otherwise False.
    """

    # standard case: all are filtered
    if allowed_list is None and denied_list is None:
        return True

    if allowed_list is not None and denied_list is not None:
        raise ValueError('in_data_filter and out_data_filter can not both be set, only one or none!')

    if allowed_list is not None:
        if not isinstance(allowed_list, list):
            allowed_list = [allowed_list]

        if name in allowed_list:
            return True

    if denied_list is not None:
        if not isinstance(denied_list, list):
            denied_list = [denied_list]

        if name not in denied_list:
            return True

    return False



def str_to_list(str, delimiter=','):
    """
    Converts a string of elements into a list.
    Example: 'ds1.var1, ds2.var2[:,-1]' -->  ['ds1.var1', 'ds2.var2[:,-1]']

    :param str: Input string.
    :param delimiter: Delimiter character used to split the string. (default=',')
    :return List with splitted string elements.
    """

    # do nothing if already a list is given
    if isinstance(str, list):
        return str

    elem_list = []

    # remove outer bracket if one exists
    str = str.strip()
    if str:
        if str[0] == '[' and str[-1] == ']':
            str = str[1:-1]

        bracket_level = 0

        cur_str = ''
        for c in str:
            if c == delimiter and bracket_level <= 0:
                elem_list.append(cur_str.strip())
                cur_str = ''
            elif c == '[':
                bracket_level += 1
                cur_str += c
            elif c == ']':
                bracket_level -= 1
                cur_str += c
            else:
                cur_str += c
        elem_list.append(cur_str.strip())

    return elem_list


def get_experiment_name():
    """Returns the name of the current experiment. Is determined based on the current working directory."""

    experiment_name = None

    # get current path
    pathstr = os.getcwd()

    # remove part in template that defines the number, for example: {:6d}
    experiment_substr = re.sub('{.*}', '', eu.EXPERIMENT_DIRECTORY_TEMPLATE)

    # search for the repetition sub_directory
    re_match = re.search('/(' + experiment_substr + '\d*)', pathstr)

    if re_match is not None:
        experiment_name = re_match.group(1)

    return experiment_name


def get_repetition_name():
    """Returns the name of the current repetition. Is determined based on the current working directory."""

    repetition_name = None

    # get current path
    pathstr = os.getcwd()

    # remove part in template that defines the number, for example: {:6d}
    repetition_substr = re.sub('{.*}', '', eu.REPETITION_DIRECTORY_TEMPLATE)

    # search for the repetition sub_directory
    re_match = re.search('/(' + repetition_substr + '\d*)', pathstr)

    if re_match is not None:
        repetition_name = re_match.group(1)

    return repetition_name


def mannwhitneyu_pvalue(data_1, data_2):

    if np.array_equal(data_1, data_2):
        return 1.0

    _, pvalue = scipy.stats.mannwhitneyu(
        data_1,
        data_2,
        alternative='two-sided')
    return pvalue


def update_status(status: str,
                  status_file: Optional[str] = None):
    """
    Updates the status of the running experiment/repetition in its status file.

    Parameters:
        status (str): Status in form of a string.
        status_file (str):
            Optional path to the status file.
            By default, it is the status file of the running process.
    """

    if status_file is None:
        status_file = os.environ.get('EU_STATUS_FILE', default=None)

    if status_file is None or status_file == '':
        warnings.warn('Can not find status file location to update its status.')
    else:
        time_str = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
        with open(status_file, 'a+') as file:
            file.write(time_str + "\n" + "running " + status + "\n")




##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu


class BaseWidget:

    @staticmethod
    def default_config():

        dc = eu.AttrDict()

        dc.is_use_state_backup = True
        dc.state_backup_name = None
        
        dc.state_backup_variable_filter = None

        return dc


    def __init__(self, config=None, **kwargs):
        self.config = eu.combine_dicts(kwargs, config, self.default_config())


    @property
    def state_backup_name(self):
        state_backup_name = self.config.state_backup_name
        if state_backup_name is None:
            state_backup_name = type(self).__name__ + '_state_backup'
        return state_backup_name


    @state_backup_name.setter
    def state_backup_name(self, state_backup_name):
        self.config.backup_name = state_backup_name


    def get_widget_state(self):
        return eu.AttrDict()


    def set_widget_state(self, state):
        pass


    def backup_state(self):
        if self.config.is_use_state_backup:
            state_backup = self.get_widget_state()
            state_backup = self._filter_state(state_backup)
            eu.gui.jupyter.save_config(state_backup, self.state_backup_name)


    def load_state_backup(self):
        if self.config.is_use_state_backup:
            state_backup = eu.gui.jupyter.load_config(self.state_backup_name)
            if state_backup:
                state_backup = self._filter_state(state_backup)
                self.set_widget_state(state_backup)
                
                
    def _filter_state(self, state):
        
        if self.config.state_backup_variable_filter is None:
            filtered_state = state
        else:
            filtered_state = eu.AttrDict()
            
            state_backup_variable_filter = self.config.state_backup_variable_filter
            if not isinstance(state_backup_variable_filter, list):
                state_backup_variable_filter = [state_backup_variable_filter]

            for variable_name in state_backup_variable_filter:
                if variable_name in state:
                    filtered_state[variable_name] = state[variable_name]
            
        return filtered_state
            
